# ---
# jupyter:
#   jupytext:
#     formats: ipynb,py:light
#     text_representation:
#       extension: .py
#       format_name: light
#       format_version: '1.3'
#       jupytext_version: 0.8.6
#   kernelspec:
#     display_name: Python 3
#     language: python
#     name: python3
# ---

from besos.parameters import RangeParameter, expand_plist, Parameter
from besos.IO_Objects import DummySelector
from besos.evaluator import EvaluatorSR
from besos.problem import Problem
from besos import optimizer

# +
parameters=[Parameter(DummySelector(), RangeParameter(0, i+1)) for i in range(3)]

def single(vals):
    return ((sum(vals),), ())

def multi(vals):
    return ((sum(vals), min(vals)),())

def multi2(vals):
    return ((sum(vals), -max(vals)), ())

single_evaluator = EvaluatorSR(single, Problem(parameters, 1))
multi_evaluator = EvaluatorSR(multi, Problem(parameters, 2,  minimize_outputs=[True, False]))
multi_min = EvaluatorSR(multi, Problem(parameters, 2,))
# -

parameters

# +
from besos.optimizer import *

single_objective = [GeneticAlgorithm]
multi_objective = [NSGAII, GDE3, SPEA2, SMPSO, CMAES,IBEA]
multi_eps = [EpsMOEA, OMOPSO, EpsNSGAII]

# the change to add JSON support somehow broke EvolutionaryStrategy and PAES

# shortcuts for periodic actions are not included
periodic_actions = [platypus.PeriodicAction, platypus.AdaptiveTimeContinuation,
                    platypus.EpsilonProgressContinuation]

# PESA2 is broken, view https://github.com/Project-Platypus/Platypus/issues/72 for details

# +
def fitness(solutions):
    for s in solutions:
        s.fitness = -s.variables[0] + s.variables[1]


def call(evaluator, alg, **kwargs):
    print(alg.__name__, end=' ')
    try:
        results = alg(evaluator, **kwargs)
        print(len(results))
    except Exception as e:
        print(e)
        
print('-- single objective')
for alg in single_objective:
    call(single_evaluator, alg)
print('-- multiple objective')
for alg in multi_objective:
    call(multi_evaluator, alg)
print('-- multiple objective, epsilon needed')
for alg in multi_eps:
    call(multi_evaluator, alg, epsilons=4)
print('-- other')
call(multi_min, MOEAD)
call(multi_min, NSGAIII, divisions_outer=4)
call(multi_evaluator, ParticleSwarm, fitness=fitness)
# -

# the results for the periodic actions are a bit odd, I may be misusing them.
for pa in periodic_actions:
    alg = platypus.NSGAII(multi_evaluator.to_platypus())
    x = pa(alg)
    x.run(1000)
    print(str(x.__class__).split('.')[-1][:-2], len(x.result))





"""
Script used to configure the inputs, outputs and defaults for Evaluators.
"""
import os
import sys
sys.path.insert(0, '..')

from pathlib import Path
import warnings
from errors import ModeError

baseDir = Path(__file__).resolve().parents[1]
# where to store output from simulations, or None for a temporary directory
out_dir = None
# where to move output files from failed runs
err_dir = Path(baseDir, 'E_Plus_Files')
# where to look for input files
data_dir = Path(baseDir, 'Data')
# whether to use 'idf' or 'json' format by default
energy_plus_mode = 'json'

# Produces a dictionary in the form {'extension': '/path/from/root/file.extension'}
# it is assumed that 'file.extension' is in the directory `inDir`
# When a file of the type 'json', 'idf', 'idd', 'epw', 'xlsx' or 'schema' is needed,
# the default is taken from this dictionary
files = {key: str(Path(data_dir, name).resolve()) for key, name in {
    'json': 'example_building.epJSON',
    'schema': 'Energy+.schema.epJSON',
    'idf': 'example_idf.idf',
    'idd': 'example_idd.idd',
    'epw': 'example_epw.epw',
    'xlsx': 'example_xlsx.xlsx',
}.items()}

# alias json and epJSON to the same building
if 'json' in files and 'epJSON'in files:
    warnings.warn('two defaults are created for json buildings')
elif 'epJSON' in files:
    files['json'] = files['epJSON']
elif 'json' in files:
    files['epJSON'] = files['json']

# set files['building'] equal to the building for the default mode
if energy_plus_mode == 'idf':
    files['building'] = files['idf']
    files['data dict'] = files['idd']
elif energy_plus_mode == 'json':
    files['building'] = files['json']
    files['data dict'] = files['schema']
else:
    raise ModeError(energy_plus_mode)

# the default min and max values for range_parameters
range_parameter = dict(
    min=0,
    max=float('inf')
)

# the default arguments for the lru_cache
cache = dict(
    enabled=True,
    size=None,
    typed=False
)

# the default objective(s) to extract from simulations
objectives = 'Electricity:Facility'  # TODO: Make this only apply to EvaluatorEP, not all evaluators

# The default `class_name` of meters/variables
objective_meter_type = 'Output:Meter'
objective_variable_type = 'Output:Variable'

# default arguments for optimizers
optimizer = dict(
    evaluations=1000,
)



# ---
# jupyter:
#   jupytext:
#     formats: ipynb,py:light
#     text_representation:
#       extension: .py
#       format_name: light
#       format_version: '1.3'
#       jupytext_version: 0.8.6
#   kernelspec:
#     display_name: Python 3
#     language: python
#     name: python3
# ---

# +
from besos import eppy_funcs as ef
from besos import sampling
from besos.evaluator import EvaluatorEP, EvaluatorSR
from besos.parameters import RangeParameter, CategoryParameter, Parameter, FieldSelector
from besos.problem import EPProblem

import pandas as pd
import copy
from subprocess import CalledProcessError
# -

# Sometimes there are parts of the design space that we want to explore that will cause the EnergyPlus simulation to fail, such as invalid combinations of parameter values. In  this example, we will sometimes try to use an undefined material `Invalid Material` in our building to represent an invalid state.

# +
building = ef.get_building()

problem = EPProblem([
    Parameter(
        FieldSelector(object_name='Mass NonRes Wall Insulation', field_name='Thickness'),
        RangeParameter(min_val = 0.01, max_val=0.99)),
    Parameter(
        FieldSelector(class_name='Construction', object_name='ext-slab', field_name='Outside Layer'),
        CategoryParameter(options=('HW CONCRETE', 'Invalid Material')))])
# -

samples = sampling.dist_sampler(sampling.lhs, problem, 5)
samples

# By default, evaluation of a DataFrame of parameters will end when an invalid combination is encountered.

try:
    EvaluatorEP(problem, building, error_mode='Failfast').df_apply(samples)
except Exception as e:
    print('caught', e)

# However sometimes we want to have a fallback value for these invalid states. This can be specified with the `error_value` argument for evaluators.  
# It must be of the form `(objective_values, constraint_values)` where objective_values and constraint_values are tuples of the same length as the number of objectives/constraints.  
#
# Since we have 1 objective and no constraints, we use a tuple with one item for objective_values, and an empty tuple for the constraint values.

# +
error_value=((10.0**20,), ())

EvaluatorEP(problem, building, error_mode='Print', error_value=error_value).df_apply(samples)
# -

# This time, the we got a warning for the invalid states, and our error value was used as their result.
#
# If we do not want to display these warnings, we can set the `error_mode='Silent'`  
#
# Ommiting the error value will use a reasonable default, set to the opposite of what we are optimizing each objective towards. (This does not work for problems with constraints)

evaluator = EvaluatorEP(problem, building, error_mode='Silent')
print('Error value defaulted to:', evaluator.error_value)

evaluator.df_apply(samples)





"""
This file is a collection of functions to interact with eppy.
They could do with some sorting and refactoring.
"""

import contextlib
import numpy as np

import config
from eppy.modeleditor import IDF
import json

from functools import partial
import warnings

from errors import ModeError

def convert_format(s: str, place, mode):
    """Converts energyPlus names to their Eppy equivalents

    :param s: name to convert
    :param place: the type of place that is being used 'field' or 'class'
    :param mode: whether to convert to idf or json formating
    :return: the converted name
    """
    # TODO: Find an authoritative source for the naming conventions
    if mode == 'idf':
        if place == 'field':
            return s.replace(' ', '_').replace('-', '')
        if place == 'class':
            return s.upper()
    if mode == 'json':
        if place == 'field':
            return s.replace(' ', '_').replace('-', '_').lower()
        if place == 'class':
            # uses camel case, s.title() removes some of the capitals and does not work
            return s
    raise ModeError(message=f'no format defined for place:{place} and mode:{mode}')


def get_mode(building):
    if isinstance(building, IDF):
        return 'idf'
    if isinstance(building, dict):
        return 'json'
    raise ModeError(message=f'Cannot find a valid mode for {building}')


def get_idf(idf_file: str = config.files.get('idf'),
            idd_file: str = config.files.get('idd'),
            output_directory=config.out_dir) -> IDF:
    """Uses eppy to read an idf file and generate the corresponding idf object"""
    # Trying to change the idd file inside a program will cause an error
    IDF.setiddname(idd_file)
    # TODO: Fix this rather than hiding it.
    # calling IDF causes a warning to appear, currently redirect_stdout hides this.
    with contextlib.redirect_stdout(None):
        idf = IDF(idf_file)
        # override the output location so I stop messing up
        idf.run = partial(idf.run, output_directory=output_directory)
        return idf


def get_building(building=None,
                 data_dict=None,
                 output_directory=config.out_dir,
                 mode=None):
    if mode is None and building is None:
        building = config.files.get('building')
        mode = config.energy_plus_mode
    if mode is None:
        ending = building.split('.')[-1]
        if ending == 'idf':
            mode = 'idf'
        elif ending == 'epJSON':
            mode = 'json'
        else:
            warnings.warn(f'expected "idf" or "epJSON" file. building has extension {ending}')
    if building is None:
        building = config.files[mode]
    if mode == 'idf':
        data_dict = data_dict or config.files['idd']
        return get_idf(idf_file=building, idd_file=data_dict, output_directory=output_directory)
    elif mode == 'json':
        with open(building) as f:
            return json.load(f)
    ModeError(mode)


def get_windows(building):
    mode = get_mode(building)
    if mode == 'idf':
        return ((window.Name, window) for window in building.idfobjects['FENESTRATIONSURFACE:DETAILED'])
    elif mode == 'json':
        return building['FenestrationSurface:Detailed'].items()
    else:
        raise ModeError(mode)


def wwr_all(building, wwr: float) -> None:
    """Sets the wwr for all walls that have a window.
    Will malfunction if there are multiple windows on one wall
    """
    mode = get_mode(building)
    windows = get_windows(building)
    for window_name, window in windows:
        wwr_single(window, wwr, building, mode)


def set_vertex(idfObj, vertexNum: int, x: float = 0, y: float = 0, z: float = 0):
    """Sets a single vertex of the passed idfObject (idfObj)
     to the specified x,y and z coordinates."""
    for val, name in zip((x, y, z), 'XYZ'):
        idfObj['Vertex_{}_{}coordinate'.format(vertexNum, name)] = round(val, 2)


def one_window(building):
    """Removes some windows so that each wall has no more than one"""
    mode = get_mode(building)
    walls = set()
    toRemove = []
    windows = get_windows(building)
    for window_name, window in windows:
        if mode == 'idf':
            wall_name = window.Building_Surface_Name
        elif mode == 'json':
            wall_name = window['building_surface_name']
        else:
            raise ModeError(mode)
        if wall_name in walls:
            toRemove.append((window_name, window))
        else:
            walls.add(wall_name)
    if mode == 'idf':
        for window_name, window in toRemove:
            building.idfobjects['FENESTRATIONSURFACE:DETAILED'].remove(window)
    elif mode == 'json':
        for window_name, window in toRemove:
            building['FenestrationSurface:Detailed'].pop(window_name)


def wwr_single(window, wwr: float, building, mode):
    """Sets `window` to have the window to wall ratio specified by `wwr`"""

    # will not work for some orientations
    # multiple windows on a single wall will break this
    if mode == 'idf':
        def coordinates(ax):
            return [window[f'Vertex_{n}_{ax.upper()}coordinate'] for n in range(1, 5)]

        wall = building.getobject('BUILDINGSURFACE:DETAILED', window.Building_Surface_Name)
    elif mode == 'json':
        def coordinates(ax):
            return [window[f'vertex_{n}_{ax.lower()}_coordinate'] for n in range(1, 5)]

        wall = building['BuildingSurface:Detailed'][window['building_surface_name']]
    else:
        raise ModeError(mode)

    xs = coordinates('X')
    ys = coordinates('Y')
    zs = coordinates('Z')

    if max(ys) == min(ys):
        axis = 'x'
        axs = xs
    elif max(xs) == min(xs):
        axis = 'y'
        axs = ys
    else:
        raise ValueError('The window is not aligned with the x or y axes')
    width = max(axs) - min(axs)
    scale_factor = np.sqrt(wwr)
    new_width = width * scale_factor
    height = max(zs) - min(zs)
    new_height = height * scale_factor

    startW = (width - new_width) / 2
    endW = startW + new_width
    startH = (height - new_height) / 2
    endH = startH + new_height

    # Maintains vertex order by mimicking the current order
    s = [0] * 4
    for vertex in range(0, 4):
        if zs[vertex] == max(zs):
            # vertex on the top
            if axs[vertex] == max(axs):
                # TOP RIGHT
                s[0] += 1
                set_vertex(window, vertex + 1, z=endH, **{axis: endW})
            else:
                # TOP LEFT
                s[1] += 1
                set_vertex(window, vertex + 1, z=endH, **{axis: startW})
        else:
            if axs[vertex] == max(axs):
                # BOTTOM RIGHT
                s[2] += 1
                set_vertex(window, vertex + 1, z=startH, **{axis: endW})
            else:
                # BOTTOM LEFT
                s[3] += 1
                set_vertex(window, vertex + 1, z=startH, **{axis: startW})
    assert s == [1] * 4, ('verticesS are wrong:', s)



# ---
# jupyter:
#   jupytext:
#     formats: ipynb,py:light
#     text_representation:
#       extension: .py
#       format_name: light
#       format_version: '1.3'
#       jupytext_version: 0.8.6
#   kernelspec:
#     display_name: Python 3
#     language: python
#     name: python3
# ---

# +
from besos import config
from besos import eppy_funcs as ef
from besos import sampling
from besos.evaluator import EvaluatorEP, AdaptiveSR
from besos.parameters import RangeParameter, expand_plist, wwr, Parameter, FieldSelector, FilterSelector
from besos.problem import EPProblem
from besos.objectives import clear_outputs

from pyKriging.krige import kriging  

import pandas as pd
# -

class KrigingEval(AdaptiveSR):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if self.problem.num_outputs != 1:
            raise ValueError('This model cannont handle multiobjective problems.')
        if self.problem.num_constraints != 0:
            raise ValueError('This model cannont handle constrained problems.')
        
    def train(self):
        self.model = kriging(self.data.values[:,:-1], self.data.values[:,-1])
        self.model.train()
        
    def eval_single(self, values):
        return ((self.model.predict(list(values)),), ())
    
    def get_infill(self, num_datapoints, **kwargs):
        return self.model.infill(num_datapoints, **kwargs)
    
    def update_model(self, new_data, old_data = None) -> None:
        for index, *row in new_data.itertuples():
            inputs = row[:self.problem.num_inputs]
            output = row[-1]
            assert len(row) == self.problem.num_inputs + self.problem.num_outputs
            self.model.addPoint(inputs, output)
        self.model.train()

# +
building = ef.get_building()
clear_outputs(building)

parameters=expand_plist(
    {
        'Mass NonRes Wall Insulation':
        {'Thickness': (0.01, 0.99)},
        'NonRes Fixed Assembly Window':
        {'U-Factor':(0.1,5),
         'Solar Heat Gain Coefficient':(0.01,0.99)},
        'Roof Membrane':
        {'Thickness': (0.003, 0.02)},
        'AtticFloor NonRes Insulation':
        {'Thickness': (0.01, 0.99)},
    })

if config.energy_plus_mode == 'idf':
    def select_zone_infiltration(idf, air_change):
        return [o for o in idf.idfobjects['ZONEINFILTRATION:DESIGNFLOWRATE']
                if (o.Air_Changes_per_Hour) == air_change]
elif config.energy_plus_mode == 'json':
    def select_zone_infiltration(building, air_change):
        return [o for name,o in building['ZoneInfiltration:DesignFlowRate'].items()
                if ('air_changes_per_hour' in o) == air_change]

parameters.extend([
    Parameter(FilterSelector(get_objects=lambda b:select_zone_infiltration(b, True),
                                 field_name='Air Changes per Hour'),
                  RangeParameter(min_val=0, max_val=1),),
    Parameter(FilterSelector(get_objects=lambda b:select_zone_infiltration(b, False),
                                 field_name='Flow per Exterior Surface Area',),
                  RangeParameter(min_val=0, max_val=0.001),
                  name='Infiltration/Area'),
    wwr(),
    Parameter(FieldSelector('Lights', '*', 'Watts per Zone Floor Area'),
                  RangeParameter(1, 20), name='Lights Watts/Area'),
    Parameter(FieldSelector('Lights', '*', 'Fraction Radiant'),
                  RangeParameter(0, 0.8)),
    # note that these fractions are a possible use case for multi-variable parameters,
    # or a valid use of error_value
    Parameter(FieldSelector('ElectricEquipment', '*', 'Watts per Zone Floor Area'),
                  RangeParameter(1, 20), name='Equipment Watts/Area'),
    Parameter(FieldSelector('ElectricEquipment', '*', 'Fraction Radiant'),
                  RangeParameter(0, 1)),
                 ])

problem = EPProblem(parameters)

evaluator = EvaluatorEP(problem, building)

train = sampling.dist_sampler(sampling.lhs, problem, 10)
train = sampling.add_extremes(train, problem)
# this helps this specific evaluator auto-detect the min/max correctly

test = sampling.dist_sampler(sampling.random, problem, 10)
answers = evaluator.df_apply(test)
# -

evaluator.df_apply(test, keep_input=True)

k = KrigingEval(reference=evaluator)
k.do_infill(train)

predictions_before = k.df_apply(test)

predictions_before

k.infill(10)

predictions_after = k.df_apply(test)

dfs = [answers, predictions_before, predictions_after]
names = ['answers', 'before', 'after']
for df, name in zip(dfs, names):
    if name not in df:
        df[name] = df['Electricity:Facility']
        del df['Electricity:Facility']
results = pd.concat(dfs, axis=1)

from sklearn.metrics import r2_score
r2_b = r2_score(results['answers'], results['before'])
r2_a = r2_score(results['answers'], results['after'])
print('before:', r2_b, 'after:', r2_a, 'change in r2:', r2_a - r2_b)

# In the past few runs, adding 10 datapoints have suggested that adding datapoints has resulted in slightly weaker models. This may be due to the model overfitting to values at the extremes of the range, but the change in r2 is very small ( <0.03)

from sklearn.metrics import r2_score
r2_b = r2_score(results['answers'], results['before'])
r2_a = r2_score(results['answers'], results['after'])
print('before:', r2_b, 'after:', r2_a, 'change in r2:', r2_a - r2_b)

results

k.data





import pytest
import pandas as pd
from pandas import DataFrame as DF
import numpy as np

from sklearn import svm, pipeline, linear_model
from sklearn.preprocessing import StandardScaler

import eppy_funcs as ef
from evaluator import EvaluatorEP, EvaluatorSR
from parameters import RangeParameter, CategoryParameter, Parameter, FieldSelector
from problem import EPProblem
import sampling

#Some code taken and adapted from example notebooks, made to work through pytest

@pytest.fixture
def building():
    return ef.get_building()


@pytest.fixture
def parameters():
    parameters = [
        Parameter(FieldSelector(object_name='NonRes Fixed Assembly Window',
                field_name='Solar Heat Gain Coefficient'),
                value_descriptor=RangeParameter(0.01,0.99)),
        Parameter(FieldSelector('Lights', '*', 'Watts per Zone Floor Area'),
                value_descriptor=RangeParameter(8, 12),
                name='Lights Watts/Area')]
    return parameters


@pytest.fixture
def problem(parameters):
    objectives = ['Electricity:Facility']
    problem = EPProblem(parameters, objectives)
    return problem


@pytest.fixture
def samples(problem):
    samples = sampling.dist_sampler(sampling.seeded_sampler, problem, 10)
    return samples


def test_expected_values(building, problem, samples):
    """check that the obtained results are consistent when using the same inputs"""

    def get_plot_data(model, density):
        #helper function from the example notebook
        p1 = problem.inputs[0].value_descriptor
        a = np.linspace(p1.min, p1.max, density)
        p2 = problem.inputs[1].value_descriptor
        b = np.linspace(p2.min, p2.max, density)
        plot_data = pd.DataFrame(np.transpose([np.tile(a, len(b)), np.repeat(b, len(a))]),
            columns=problem.names('inputs'))
        return pd.concat([plot_data, pd.Series(model.predict(plot_data))], axis=1)

    evaluator = EvaluatorEP(problem, building, error_mode='Silent')
    train = evaluator.df_apply(samples, keep_input=True)
    print(problem.names())
    x,y,c = problem.names()

    #train and get the values
    model = pipeline.make_pipeline(
        StandardScaler(),
        linear_model.Ridge())
    model.fit(train[[x, y]].values, train[c].values)
    density = 30
    df = get_plot_data(model, int(density*1.5))

    #check to see if the extremes and the midpoint are the expected values
    assert np.isclose([df.iloc[0][0]], [1592469368.6909447]), f'Unexpected value for low extreme'
    assert np.isclose([df.iloc[2024][0]], [2122632827.9627075]), f'Unexpected value for high extreme'
    assert np.isclose([df.iloc[1012][0]], [1857551098.326826]), f'Unexpected value for midpoint'

    #change this to 0 to see stdout and stderr
    assert 1



import pyDOE2
import numpy as np
import pandas as pd

from typing import Callable, Any

from IO_Objects import Descriptor
from problem import Problem



########################################################################################################
##############################     Static  sampling     ################################################
########################################################################################################


_sampler = Callable[[int, int, Any], np.ndarray]  # the format of all sampler functions


def dist_sampler(sampler: _sampler, problem: Problem, num_samples: int, *args, **kwargs) -> pd.DataFrame:
    """uses the sampling function provided to generate `num_samples` sets of values
    These values will be valid for the corresponding inputs

    :param problem: Problem that the inputs should apply to
    :param sampler: a function that is used to produce the distribution of samples
    :param num_samples: number of samples to take

    :param args: Arguments passed to the sampling function
    :param kwargs: Arguments passed to the sampling function
    :return: pandas DataFrame with one column per parameter, and `num_samples` rows

    """
    samples = sampler(num_samples, problem.num_inputs, *args, **kwargs)
    data = {name: _sample_param(param, samples[:, i]) for i, (name, param)
            in enumerate(zip(problem.names('inputs'), problem.inputs))}
    df = pd.DataFrame(data)
    # enforce the correct order in case it was lost by the dictionary
    df = df[problem.names('inputs')]
    return df


def add_extremes(df, problem):
    """adds two datapoints with the minimum and maximum values for each attribute

    :param data:
    :param problem:
    :return:
    """
    new = dist_sampler(extremes, problem, 2)
    return pd.concat([df, new], ignore_index=True)


def extremes(samples: int, attributes: int=None) -> np.ndarray:
    if attributes is None:
        samples, attributes = 2, samples
    assert samples == 2, "extremes can only produce two samples"
    return np.array([np.zeros(attributes), np.ones(attributes)])


def _sample_param(parameter: Descriptor, values):
    try:
        # this will only work if p.sample is numpy compatible
        return parameter.sample(values)
    except TypeError:
        # apply non-vectorised version of p.sample
        return [parameter.sample(x) for x in values]


# all of the following return a 2d array of shape (samples, attributes)
# which contains values between 0 and 1


random: _sampler = np.random.rand


def seeded_sampler(samples: int, attributes: int, seed=0) -> np.ndarray:
    np.random.seed(seed)
    return np.random.rand(samples, attributes)


def lhs(samples: int, attributes: int, *args, **kwargs) -> np.ndarray:
    return pyDOE2.lhs(attributes, samples, *args, **kwargs)

def full_factorial(samples: int, attributes: int, level: int=None) ->np.ndarray:
   if level == None:
       level = np.int(np.exp(np.log(samples)/attributes))
       if (level**attributes) > level:
        print(f'Total number of samples ({level**attributes}) is smaller  than input ({samples}) '
              f'to have an even number of factor levels ({level}) for all parameters.')
   return pyDOE2.fullfact((np.ones([attributes]) * level).astype(int))/level





########################################################################################################
##############################    Adaptive  sampling    ################################################
########################################################################################################


class adaptive_sampler_lv():
    """Represents an adaptive sampler which iteratively picks simulation samples and runs them."""
    def __init__(self, P, P_out, n_new, problem, evaluator, reg, scaler=None, scaler_out=None, verbose = False):
        if len(P[:,0])<2*len(P[0,:]):
            print('Size of initial number of samples needs to be bigger than 2 times no. inputs.')
        self.P = P
        self.P_out = P_out
        self.n_new = n_new
        self.problem = problem
        self.evaluator = evaluator
        self.reg = reg
        self.scaler = scaler
        self.scaler_out = scaler_out
        self.verbose = verbose

    def run(self, no_iter):

        self.score = np.empty([no_iter+1])

        self.N_P, self.S_P = init_Neighborhood(self.P)
        self.pick_new_samples()
        self.update_model()

        for i in range(no_iter):
            print(i)
            for p_new in self.P_new:
                self.N_P, self.S_P = update_Neighborhood(self.N_P, self.P, self.S_P, p_new)
                if self.verbose==True:
                    print('initialize neighborhood for new samples')
                N_P_new, S_P_new = init_Neighborhood(self.P, p_new)
                self.N_P = np.append(self.N_P, N_P_new, axis=2)
                self.S_P = np.append(self.S_P, S_P_new, axis=0)
            # update existing set of samples
            self.P = np.append(self.P, self.P_new, axis=0)
            self.P_out = np.append(self.P_out, self.P_out_new, axis=0)
            self.pick_new_samples()

            self.update_model()

            self.score[i] = self.reg.score(self.P, self.P_out)

    def update_model(self):
        self.reg.fit(self.P, self.P_out)
        if self.verbose == True:
            print(self.reg.score(self.P, self.P_out))

    def pick_new_samples(self):
        #2) Compute nonlinearity measure and Voronoi Cell
        E = LOLA_estimate(self.N_P, self.P, self.reg, scaler=self.scaler, scaler_out=self.scaler_out)
        V_P,samples = MC_Voronoi_estimate(self.P, self.problem)
        H = hybrid_score(E, V_P)
        ind_new = np.argsort(H)
        P_sorted = self.P[ind_new,:]
        if self.verbose==True:
            print('Collecting new samples around:')
            print(P_sorted[-self.n_new:,:])

        # Generate new samples in Voronoi cell taking the mean of the neighborhood
        ind=2
        self.P_new=np.empty([self.n_new,len(self.P[0,:])])
        for i in range(self.n_new):
            candidates = containedin_Voronoi(P_sorted[-i-1, :], self.P, samples)
            #print(len(candidates))
            while len(candidates)<=1:
                #print('in loop')
                candidates = containedin_Voronoi(P_sorted[-i - ind, :], self.P, samples)
                ind+=1
            self.P_new[i,:] = select_newSamp(P_sorted[-i-1, :], self.N_P[:,:,ind_new][:,:,-i-1], candidates)
        self.P_out_new = self.evaluator.df_apply(pd.DataFrame(self.P_new))




###############################################################
def hybrid_score(E,V):
    return V+E/sum(E)

def select_newSamp(p_r, N, candidates, verbose=False):
    N = np.append(N, [p_r], axis=0)
    max_dist = 0
    if verbose == True:
        print(len(candidates[1:, :]))
    for c in candidates[1:,:]:
        d=0
        for n in N:
            d = d + np.linalg.norm(c-n)
        if d>max_dist:
                max_dist = d
                max_cand = c
    return max_cand



def MC_Voronoi_estimate(P, problem):
    # P are the existing points

    # initialize arrays
    V_P = np.zeros([len(P),])  # Vector of all Voronoi sample estimates
    # per existing point collect randomly selected 100 points

    S = dist_sampler(sampling.lhs, problem, 500) # for now 100 random samples per point to estimate Voronoi cell

    for s in S.values:
        d = np.inf
        ind = 0
        for p in P:
            if np.linalg.norm(p-s) < d:
                d = np.linalg.norm(p-s)
                ind_fin = ind
            ind = ind + 1
        V_P[ind_fin] = V_P[ind_fin] + 1/len(S)
    #plt.plot(S.values[:, 0], S.values[:, 1], 'x')
    return V_P,S


def LOLA_estimate(N_P, P, model, scaler, scaler_out):
    # N_P is a 3-D array with all neighbors for each sample (mxdx#samples)
    # S_P is a 1_D array with the neighborhood score for all samples
    # P is the array including all samples

    n = len(P)
    ind = 0
    E = np.empty([n,])
    for p_r in P:
        grad = Gradient_estimation(N_P[:, :, ind], p_r, model, scaler=scaler, scaler_out=scaler_out)
        E[ind] = Nonlinearity_measure(grad, N_P[:, :, ind], p_r, model, scaler=scaler, scaler_out=scaler_out)
        ind=ind+1

    return E

def update_Neighborhood(N_P, P, S_P , P_new):
    # this function tries to add one new sample p_new to all of the existing neighborhoods
    # S is the neighborhood score
    # N_P are the associated neighbors of all points in P
    # P is the matrix of reference points
    # p_new the candidate
    if np.ndim(P)==1:
        P = np.expand_dims(P,axis=0)
        N_P = np.reshape(N_P,(len(N_P[:,0]), len(N_P[0,:]), 1))
        S_P = np.expand_dims(S_P, axis = 0)

    if np.ndim(P_new) == 1:
        P_new = np.expand_dims(P_new,axis=0)

    m = 2 * len(P[0,:])
    ind = 0
    for p_new in P_new:
        for p_r in P:
            if sum(p_r == p_new) <len(P[0,:]):
                N_temp = np.dstack([N_P[:, :, ind]]*m)
                S_temp = np.zeros([m,])
                for i in range(m):
                    # replace one sample with the new one
                    if sum(sum(N_temp[:,:,i]==p_new))<len(P[0,:]):
                        N_temp[i, :, i] = p_new
                    else:
                        pass
                    # compute neighborhood score for all possible combinations
                    S_temp[i] = Neighborhood_score(N_temp[:, :, i], p_r)
                    # if neighborhood score is better than the score without the new sample, update the neighborhood
                ind_min = np.argmin(S_temp)
                #print(ind_min)
                #print(S_temp)
                N_P[:,:,ind]=N_temp[:,:,ind_min]
                S_P[ind] = S_temp[ind_min]
                ind = ind + 1
            else:
                pass
                #print(N_P[:,:,1])
    return N_P, S_P

def init_Neighborhood(P, P_new=None, verbose=False):
    # P is the matrix with all samples
    # pnew optional if the neighborhood for new samples shall be computed
    # for now dim(pnew)=1 is assumed
    m = 2 * len(P[0,:]) # number of samples in neighborhood
    d = len(P[0,:])# number of dimensions

    if np.any(P_new==None):
        # case I: first initialization after static sampling
        n = len(P)  # number of samples
        N_P = np.empty([m,d,n])
        S_P = np.empty([n,])
        P_ref = P*1.0   # array of all samples to be considered as reference in the next for loop
    else:
        # case II: initialize neighborhoods of new samples p_new
        if np.ndim(P_new) == 1:
            if verbose==True:
                print('ndim')
            P_ref = np.expand_dims(P_new, axis=0)
        else:
            P_ref = P_new*1.0
        n_new = len(P_ref)  # number of new samples
        N_P = np.empty([m,d,n_new])
        S_P = np.empty([n_new,])


    for i in range(len(P_ref)):
        mask = P != P_ref[i,:]
        P_NoRefPoint = P[mask[:, 0], :]
        N_P[:,:,i] = P_NoRefPoint[0:m, :]
        S_P[i] = Neighborhood_score(N_P[:,:,i], P_ref[i,:])
    ind = 0
    for p_cand in P:
        ind +=1
        # at the moment p_new to in P --> only 1 p_new allowed
        N_P, S_P = update_Neighborhood(N_P, P_ref, S_P, p_cand)
    if verbose == True:
        print('Neighborhood initiated!')
    return N_P, S_P


def Neighborhood_score(N, p_r):
    # To Do: make it possible to have 3D arrays as inputs
    # N is the neighborhood
    # P_cand should have m entries
    m = len(N)

    # initialize arrays
    Dist_cand = np.empty([m,])
    minDist = np.empty([m, ])

    # compute distance matrix
    C = 0
    for i in range(m):
        C=C+np.linalg.norm(N[i,:] - p_r)
    C = C/m

    for i in range(m):
        for j in range(m):
            if i==j:
                pass
            else:
                Dist_cand[i] = np.linalg.norm(N[i,:]-N[j,:])
        minDist[i] = min(Dist_cand)

    A = 1/m*sum(minDist)
    R = A/(np.sqrt(2)*C)
    return R/C

def Gradient_estimation(N, p_r, model,scaler, scaler_out):
    m = len(N)
    d = len(p_r)

    P_mat = np.empty([m, d])
    F_mat = np.empty([m,])

    for i in range(m):
            P_mat[i, :] = N[i,:]-p_r
            if scaler == None:
                F_mat[i] = model.predict([N[i,:]])
            else:
                F_mat[i] = scaler_out.inverse_transform(model.predict(scaler.transform([N[i, :]])))
    #grad = np.linalg.solve(P_mat,F_mat.reshape((len(F_mat),1)))
    grad = np.linalg.lstsq(P_mat, np.transpose(F_mat),rcond=None)[0].reshape((1,d))
    return [grad]

def Nonlinearity_measure(grad, N, p_r, model, scaler, scaler_out):
    E=0
    for i in range(len(N)):
        if scaler == None:
            E = E + abs(model.predict([N[i, :]]) -
                        (model.predict([p_r]) + np.dot(grad,(N[i, :]-p_r))))
        else:
            E = E + abs(scaler_out.inverse_transform(model.predict(scaler.transform([N[i, :]]))) -
                        (scaler_out.inverse_transform(model.predict(scaler.transform([p_r]))) + np.dot(grad,(N[i, :]-p_r))))
    return E



def containedin_Voronoi(p_r,P,samples):
    mask = P != p_r
    P_temp = P[mask[:, 0], :]
    candidates=np.empty([1,len(P[0,:])])
    for s in samples.values:
        for p_j in P_temp:
            if np.linalg.norm(s-p_j)<=np.linalg.norm(s-p_r):
                break
            else:
                continue
        if np.all(p_j == P[-1]):
            candidates = np.append(candidates, [s], axis=0)

    return candidates



import pytest

import eppy_funcs as ef
from evaluator import EvaluatorEP, EvaluatorSR
from parameters import RangeParameter, CategoryParameter, Parameter, FieldSelector, FilterSelector, GenericSelector, wwr
from problem import EPProblem, Problem
from optimizer import NSGAII
from objectives import MeterReader
import sampling


@pytest.fixture
def building():
    building = ef.get_building(mode='json')
    # we need the json example file because of how the insulation_filter function works


def test_init(building):
    """to make sure each type of selector is created as expected"""

    def insulation_filter(building):
        return [obj for name, obj in building['Material'].items() if 'Insulation' in name]

    field_selector = FieldSelector(class_name='Material', object_name='Mass NonRes Wall Insulation', field_name='Thickness')
    filter_selector = FilterSelector(insulation_filter, 'Thickness')
    generic_selector = GenericSelector(set=ef.wwr_all, setup=ef.one_window)

    #basic checks for the expected output format
    assert str(field_selector) == "FieldSelector(field_name='Thickness', class_name='Material', object_name='Mass NonRes Wall Insulation')", f'FieldSelector not properly initialised: {field_selector}'
    assert filter_selector.field_name == 'Thickness' and str(type(filter_selector.get_objects)) =="<class 'method'>", f'FilterSelector not properly initialised: {filter_selector}'
    assert str(type(generic_selector.set)) =="<class 'method'>" and str(type(generic_selector.setup)) =="<class 'method'>", f'GenericSelector not properly initialised: {generic_selector}'

    #change this to 0 to see stdout and stderr
    assert 1



"""
This script containts a custom error created to avoid code repetion for mode check in eppy_funcs and config.
"""
class ModeError(ValueError):
    def __init__(self, mode=None, message=None):
        if message is None:
            message = f'Invalid mode {mode}. Expected "idf" or "json"'
        super().__init__(message)



# ---
# jupyter:
#   jupytext:
#     formats: ipynb,py:light
#     text_representation:
#       extension: .py
#       format_name: light
#       format_version: '1.3'
#       jupytext_version: 0.8.6
#   kernelspec:
#     display_name: Python 3
#     language: python
#     name: python3
# ---

# # Descriptors

# +
from besos.parameters import RangeParameter, CategoryParameter, Parameter
from besos.problem import Problem
from besos import sampling
from besos.evaluator import EvaluatorSR 


import pandas as pd
# -

# Descriptors specify what kinds of values are valid for a parameter. Currently, two variants exist: `RangeParameter` and `CategoryParameter`.  
#
# RangeParameters have a minimum and a maximum, and allow any value in that range. The endpoints are included in the allowable range.

zero_to_nine = RangeParameter(min_val = 0, max_val=9)

# CategoryParameters have a list of options, and can only be set to the value of items from that list.

single_digit_integers = CategoryParameter(options=[0,1,2,3,4,5,6,7,8,9])
text_example = CategoryParameter(options=['a', 'b', 'c', 'other'])

# ### Sampling
# Parameters only need the Descriptor part for sampling.

# +
parameters = [
    Parameter(value_descriptor=zero_to_nine, name='0-9'),
    Parameter(value_descriptor=single_digit_integers, name='single digit'),
    Parameter(value_descriptor=text_example, name='text')
]
problem = Problem(parameters, outputs=['output'])

samples = sampling.dist_sampler(sampling.lhs, problem, num_samples=10)
samples
# -

# ### Evaluation
# Since we did not specify selectors for the parameters, we cannot evaluate them using an energy plus simulation.  
# Instead, we will use a custom evaluation function.

# +
def evaluation_function(values):
    x,y,z = values
    if z == 'other':
        return (0,), ()
    else:
        return (x * y,), ()


evaluator = EvaluatorSR(evaluation_function, problem)
# The evaluator will use this objective by default
outputs = evaluator.df_apply(samples ,keep_input=True)
# outputs is a pandas dataframe with one column since only one objective was requested
# -

outputs





import os
import platypus
import rbfopt
import pandas as pd
import numpy as np
from contextlib import redirect_stdout
import sys

from evaluator import AbstractEvaluator
from config import optimizer as conf

from typing import Type, List
from inspect import signature, Parameter, Signature
from functools import partial
from collections import OrderedDict


# TODO: evaluations or end criterion?
def platypus_alg(evaluator: AbstractEvaluator,
                 algorithm: Type[platypus.Algorithm], evaluations: int = conf['evaluations'],
                 *args, **kwargs) -> pd.DataFrame:
    """Uses a platypus algorithm to optimise over an evaluator

    :param evaluator: An evaluation function to optimise over.
    :param algorithm: The platypus algorithm to use.
    :param evaluations: The algorithm will be stopped once it uses more than this many evaluations.
    :param args: arguments to pass to `algorithm`
    :param kwargs: keyword arguments to pass to `algorithm`.
    :return: the non-dominated solutions found by the algorithm.
    """

    problem = evaluator.to_platypus()
    alg: platypus.AbstractGeneticAlgorithm = algorithm(problem, *args, **kwargs)
    alg.run(evaluations)
    # TODO: Have a smarter default for parts (instead of 'all')
    return solutions_to_df(platypus.unique(alg.result), evaluator.problem)


_ordered_argtypes = [(Parameter.POSITIONAL_OR_KEYWORD, False),
                     (Parameter.POSITIONAL_OR_KEYWORD, True),
                     (Parameter.VAR_POSITIONAL, False),
                     (Parameter.KEYWORD_ONLY, False),
                     (Parameter.KEYWORD_ONLY, True),
                     (Parameter.VAR_KEYWORD, False)]


def get_operator(problem: platypus.Problem, mutation=False):
    """Creates a valid operator for the given platypus Problem.
    This is provided to help when optimizing over problems with mixed types.

    :param problem: the platypus problem this operator should apply to
    :param mutation: if True create a mutation operator, if False create a variation operator
    :return:
    """
    operators = []
    if mutation:
        defaults = platypus.config.PlatypusConfig.default_mutator
        class_ = platypus.CompoundMutation
    else:
        defaults = platypus.config.PlatypusConfig.default_variator
        class_ = platypus.CompoundOperator
    for t in problem.types:
        if t in defaults:
            operators.append(defaults[t])
        else:
            for super_t, operator in defaults.items():
                if isinstance(t, super_t):
                    operators.append(operator)
    if mutation:
        return class_(*operators)
    return class_(*operators)


def _alg_t(algorithm):
    """A function that wraps a platypus algorithm in a common workflow.
    This function propagates the algorithm's function signature out to produce a more descriptive
    function signature. """

    # This is perhaps an overly fancy way of doing it, but it avoids having to
    # do the same process manually for all platypus algorithms, and will
    # update the signatures automatically if they change in platypus.

    defaults = {'variator': get_operator,
                'mutator': partial(get_operator, mutation=True)}

    def template(evaluator, evaluations=conf['evaluations'], *args, **kwargs):
        for kwarg, default in defaults.items():
            if kwarg not in kwargs or kwargs[kwarg] == 'automatic':
                kwargs[kwarg] = default(evaluator.to_platypus())
        return platypus_alg(evaluator, algorithm, evaluations, *args, **kwargs)

    temp_name = (
        # (function, arguments_to_omit)
        (template, ['args', 'kwargs']),
        (algorithm.__init__, ['self', 'problem'])
    )

    argdict = {kind: [] for kind in _ordered_argtypes}
    for i, (func, omit) in enumerate(temp_name):
        sig = signature(func)
        p: Parameter
        for name, p in sig.parameters.items():
            if name in omit:
                continue
            if i > 0:
                # convert non-default keyword arguments to keyword only arguments
                if p.kind is Parameter.POSITIONAL_OR_KEYWORD and p.default is Parameter.empty:
                    p = p.replace(kind=Parameter.KEYWORD_ONLY)
                # use the defaults specified above
                if name in defaults:
                    p = p.replace(default='automatic')

            argdict[(p.kind, p.default is not Parameter.empty)].append(p)
    arglist = []
    for kind, default in _ordered_argtypes:
        arglist.extend(argdict[kind, default])

    template.__signature__ = Signature(parameters=arglist,
                                       return_annotation=signature(template).return_annotation)
    template.__name__ = algorithm.__name__
    return template


# The following functions are convenience wrappers for platypus' implementations
# of the algorithms with the same name
# they return the non-dominated solutions found by the algorithm

GeneticAlgorithm = _alg_t(platypus.GeneticAlgorithm)
EvolutionaryStrategy = _alg_t(platypus.EvolutionaryStrategy)
NSGAII = _alg_t(platypus.NSGAII)
EpsMOEA = _alg_t(platypus.EpsMOEA)
GDE3 = _alg_t(platypus.GDE3)
SPEA2 = _alg_t(platypus.SPEA2)
MOEAD = _alg_t(platypus.MOEAD)
NSGAIII = _alg_t(platypus.NSGAIII)
ParticleSwarm = _alg_t(platypus.ParticleSwarm)
OMOPSO = _alg_t(platypus.OMOPSO)
SMPSO = _alg_t(platypus.SMPSO)
CMAES = _alg_t(platypus.CMAES)
IBEA = _alg_t(platypus.IBEA)
PAES = _alg_t(platypus.PAES)
# PESA2 = _alg_t(platypus.PESA2)  # implementation is broken: github.com/Project-Platypus/Platypus/issues/72
EpsNSGAII = _alg_t(platypus.EpsNSGAII)


def values_to_solution(values: list, problem: platypus.Problem, evaluate=True) -> platypus.Solution:
    """ Produces an unevaluated platypus solution with the given inputs

    :param values: a list of input values to convert to a platypus solution
    :param problem: the platypus problem this solution applies to
    :param evaluate: whether to evaluate the solution before returning
    :return: the corresponding platypus.Solution object
    """
    solution = platypus.Solution(problem)
    if len(values) != problem.nvars:
        raise ValueError(f'Length of values is {len(values)} but expected length of {problem.nvars}')
    solution.variables = list(values)
    if evaluate:
        solution.evaluate()
    return solution


# TODO: Unify `parts` handling between this and Problem to remove duplicate code if possible

def solution_to_values(solution: platypus.Solution,
                       parts='all') -> list:
    """Converts a platypus solution to a list containing the same values

    :param solution: a platypus solution to convert
    :param parts: which parts of the solution should be kept
    :return: a list of the requested values from the solution
    """
    part_options = OrderedDict(inputs='variables', outputs='objectives',
                               constraints='constraints', violation='violation')
    if parts == 'all':
        parts = part_options.keys()
    assert set(parts) <= set(part_options), f'parts must be a subset of {set(part_options.keys())} or "all"'
    result = []
    for part in parts:
        if part == 'violation':
            result.append(solution.constraint_violation)
        elif part == 'inputs':
            encoded = getattr(solution, part_options[part])
            decoded = [plat_type.decode(val) for val, plat_type in zip(encoded, solution.problem.types)]
            result.extend(decoded)
        else:
            result.extend(getattr(solution, part_options[part]))
    return result


def solutions_to_df(solutions: List[platypus.Solution], problem, parts='all', flag_optimal=True) -> pd.DataFrame:
    """Converts a list of platypus solutions to a DataFrame, with one row corresponding to each solution

    :param solutions: list of solutions to convert
    :param problem: the column names for DataFrame
    :param parts: which parts of the solutions should be kept
    :param flag_optimal: whether to include a boolean column denoting whether each solution is pareto-optimal
    :return: a DataFrame
    """

    def to_col_vals(solution_list):
        return list(zip(*(solution_to_values(solution, parts) for solution in solution_list)))

    solutions = platypus.unique(solutions)
    non_dominated = platypus.nondominated(solutions)
    columns = problem.names(parts)
    values, non_dom_vals = to_col_vals(solutions), to_col_vals(non_dominated)
    assert len(columns) == len(values), f'{len(values)} values does not match {len(columns)} columns'
    # TODO: Intuit the dataframe column types based on the types of the parameters of the problem
    # or use the to_df method of the problem object
    solution_df = pd.DataFrame({column: data for column, data in zip(columns, values)})  # , dtype=float
    if flag_optimal:
        non_dom_df = pd.DataFrame({column: data for column, data in zip(columns, non_dom_vals)})  # , dtype=float
        df = pd.merge(solution_df, non_dom_df, how='outer', indicator='pareto-optimal')
        df['pareto-optimal'] = df['pareto-optimal'] == 'both'
        return df
    return solution_df


def df_to_solutions(df: pd.DataFrame, problem: platypus.Problem) -> List[platypus.Solution]:
    """Converts a DataFrame to a list of platypus solutions, with each row converted to one solution

    :param df: DataFrame to convert
    :param problem: platypus problem that the solutions apply to
    :return: a list of platypus solutions
    """
    return list(df.apply(values_to_solution, problem=problem, axis=1))


# rbf opt

# TODO: Add support for the other termination criterion
def rbf_opt(evaluator, evaluations, hide_output=True):
    problem = evaluator.problem
    inputs = problem.inputs
    dimension = problem.num_inputs
    num_outputs = problem.num_outputs
    assert num_outputs == 1, f'This is a single objective optimizer, but you have {num_outputs} objectives'
    assert problem.num_constraints == 0, 'This optimizer cannot handle constraints'
    # may throw an attribute error if invalid parameters are used (i.e categorical)
    param_attrs = [(p.value_descriptor.min, p.value_descriptor.max, p.value_descriptor.rbf_type)
                   for p in inputs]
    var_lower, var_upper, var_type = (np.array(x) for x in zip(*param_attrs))
    # with the current classes var_type is just an array of 'R's
    # this would change if we add a descriptor for integers
    print(var_lower, var_upper, var_type)

    def wrapper(values):
        return evaluator(values)[0]

    def run_alg():
        bb = rbfopt.RbfoptUserBlackBox(dimension=dimension, var_lower=var_lower,
                                       var_upper=var_upper, var_type=var_type, obj_funct=wrapper)
        settings = rbfopt.RbfoptSettings(max_evaluations=evaluations)
        alg = rbfopt.RbfoptAlgorithm(settings, bb)
        return alg.optimize()

    if hide_output:
        with open(os.devnull, 'w') as null:
            old, sys.stdout = sys.stdout, null
            try:
                val, x, itercount, evalcount, fast_evalcount = run_alg()
            finally:
                sys.stdout = old
    else:
        val, x, itercount, evalcount, fast_evalcount = run_alg()


    return problem.to_df(np.array([np.concatenate([x, [val]])]))



# ---
# jupyter:
#   jupytext:
#     formats: ipynb,py:light
#     text_representation:
#       extension: .py
#       format_name: light
#       format_version: '1.3'
#       jupytext_version: 0.8.6
#   kernelspec:
#     display_name: Python 3
#     language: python
#     name: python3
# ---

# +
from besos import eppy_funcs as ef

from besos.parameters import RangeParameter, expand_plist, wwr
from besos.objectives import clear_outputs
from besos.evaluator import EvaluatorEP
from besos.problem import EPProblem
from besos import optimizer
import platypus
from besos import sampling
import pandas as pd

import colorsys
from pandas.plotting import parallel_coordinates
import matplotlib.pyplot as plt
import seaborn as sns
# %matplotlib inline

# +
building = ef.get_building()
clear_outputs(building)

parameters = expand_plist(
    #{'Name of object in idf':
    # {'Property Name':(min, max)}}
    {'NonRes Fixed Assembly Window':
     {'U-Factor':(0.1,5),
      'Solar Heat Gain Coefficient':(0.01,0.99)
     },
     'Mass NonRes Wall Insulation':{'Thickness':(0.01,0.09)},
    })
parameters.append(wwr())

objectives = ['Cooling:EnergyTransfer', 'Heating:EnergyTransfer', 'Electricity:Facility']

constraints = ['CO2:Facility']

problem = EPProblem(parameters, objectives, constraints, constraint_bounds=['<= 900'])

evaluator = EvaluatorEP(problem, building)

plat_problem = evaluator.to_platypus()
# -

'FENESTRATIONSURFACE:DETAILED'.title()

class LoggingArchive(platypus.Archive):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, *kwargs)
        self.log = []
    
    def add(self, solution):
        super().add(solution)
        self.log.append(solution)

archive = LoggingArchive()
solutions = optimizer.NSGAII(evaluator, population_size=10, archive=archive, evaluations=20)
df = optimizer.solutions_to_df(archive.log, problem=problem)

df

# +
fig = plt.figure(figsize=(6,6))
ax = fig.add_subplot(111)

a = df['pareto-optimal']
b = df['violation'] == 0

x_label = 'Cooling:EnergyTransfer'
y_label = 'Heating:EnergyTransfer'

for optimal, valid in ((False, False), (False, True), (True, True)):
    if optimal:
        colour = 'g'
    else:
        colour = 'b'
    if valid:
        marker = 'o'
    else:
        marker = 'x'
    mask = (a==optimal) & (b==valid)
    d = df[mask]
    ax.scatter(x=d[x_label], y=d[y_label], marker=marker,
               c=colour, label=f'opt:{optimal}, valid:{valid}')

ax.legend()
ax.set_xlabel(x_label)
ax.set_ylabel(y_label)
# ax.set_ylim(0,2*10**9)
# ax.set_xlim(0,2*10**9)
# -

sns.pairplot(df, vars=problem.names(('inputs', 'outputs', 'constraints')), hue='pareto-optimal')

# adapted from https://stackoverflow.com/q/34071476/8061009
def heatmap(df, problem, parts=('inputs', 'outputs'), sort_by=None, cspace=2, figsize=(6,6)):
    if isinstance(parts, str):
        parts = (parts,)
    columns = problem.names(parts)
    if sort_by:
        df = df.sort_values(sort_by)
    df = df[columns]
    part_map = {part: {'names':problem.names(part)} for part in parts}
    num_colours = sum(len(v['names']) for v in part_map.values()) + cspace * (len(part_map)-1)
    for v in part_map.values():
        v['colours'] = [colorsys.hsv_to_rgb(x*1.0/(num_colours), 0.7, 0.5) for x in range(len(v['names']))]
    
    fig, ax = plt.subplots(figsize=figsize)

    with sns.axes_style('white'):
        for v in part_map.values():
            for colors, name in zip(v['colours'], v['names']):
                # Create cmap
                cmap = sns.light_palette(colors, input='rgb', reverse=False, as_cmap=True)

                sns.heatmap(df.mask(df.isin(df[name])!=1),
                            ax=ax,
                            cbar=False,
                            square=True,
                            annot=False,
                            cmap=cmap,
                            linewidths=0.1)
    ax.yaxis.set_visible(False)
    plt.show()

heatmap(df[df['pareto-optimal']], problem, sort_by='Heating:EnergyTransfer')

def par_coodrs(df, selector=None, label='selected', filter_=None, default_label='no_label',
              figsize=(10,8), columns=None, limit=None):
    fig, ax = plt.subplots(figsize=figsize) 
    for column in df:
        if df[column].min() == df[column].max():
            if columns is not None and column in columns:
                columns.remove(column)
                print(column, 'has a constant value of', df[column].min())
            df = df.drop(column, axis=1)
            
    
    df = (df - df.min())/(df.max()-df.min())
    
    df['colour'] = default_label
    if selector is not None:
        df.loc[selector, 'colour'] = label
    if filter_ is not None:
        df = df[filter_]
    if columns:
        df = df[columns + ['colour']]
    if limit is not None:
        df = df.sample(limit)
    parallel_coordinates(df, 'colour', colormap='prism', ax=ax)
    ax.yaxis.set_visible(False)
    plt.show()

columns = problem.names(('inputs', 'outputs', 'constraints'))
columns

par_coodrs(df,
           selector=(df['violation'] == 0) & (df['pareto-optimal']),
           columns=problem.names(('inputs', 'outputs', 'constraints')))





from typing import Union
from abc import ABC, abstractmethod


class ReprMixin(ABC):
    """Adds a useful string representation to subclasses"""
    def __init__(self):
        self._to_repr = {}

    # TODO: replace all _to_repr[name] = attr lines with these shortcuts
    def _add_repr(self, name, attr=None, check=False):
        if attr is None:
            attr = name
        if (not check) or getattr(self, attr):
            self._to_repr[name] = attr

    def _add_reprs(self, names, check=False):
        for name in names:
            self._add_repr(name, check=check)

    def __repr__(self):
        attr_list = [f'{attr_name}={repr(getattr(self, attr))}'
                     for attr_name, attr in self._to_repr.items()]
        return f'{self.__class__.__name__}({", ".join(attr_list)})'


class IOBase(ReprMixin, ABC):
    """Base class for inputs, objectives and constraints.

    `name` is used for generating DataFrame column titles."""

    def __init__(self, name=''):
        super().__init__()
        self._name = name
        if name:
            self._to_repr['name'] = '_name'

    @property
    def name(self):
        return self._name or self._default_name

    @name.setter
    def name(self, value):
        self._name = value

    @property
    def _default_name(self):
        return ''


def get_name(x: Union[str, int, IOBase]):
    if isinstance(x, str):
        return x
    if isinstance(x, int):
        return str(x)
    return x.name


class Descriptor(ReprMixin, ABC):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.platypus_type = NotImplemented

    @abstractmethod
    def validate(self, value) -> bool:
        """Checks if value is a valid value for this parameter.

        :param value:
        :return: True if the value is valid False otherwise
        """

    # consider a pandas type to improve DataFrame generation
    @abstractmethod
    def sample(self, value):
        """Takes a value in the range 0-1 and returns a valid value for this parameter"""
        pass


class AnyValue(Descriptor):
    """A descriptor that can take on any possible value.
    Intended to be used as a placeholder."""
    def validate(self, value):
        return True

    def sample(self, value):
        raise NotImplementedError

    def __bool__(self):
        """Marks this as a 'missing' piece"""
        # ReprMixin will detect this as something to not include when checking
        return False


class Selector(ReprMixin, ABC):
    """Base Class for Selectors, which describe what attribute of the building is read or modified"""
    @abstractmethod
    def get(self, building):
        pass

    @abstractmethod
    def set(self, building, value):
        pass

    def setup(self, building) -> None:
        """Modifies the building so that it is ready for this selector"""
        pass


class DummySelector(Selector):
    """A selector that does not modify the building.
    Intended to be used as a placeholder."""
    def get(self, building):
        raise NotImplementedError

    def set(self, building, value):
        pass

    def __bool__(self):
        """Marks this as a 'missing' piece"""
        # ReprMixin will detect this as something to not include when checking
        return False


class Objective(IOBase):
    """Base class for objectives."""

    def __call__(self, *args, **kwargs) -> float:
        """Return the objective's value on the instance represented by *args and **kwargs"""
        raise NotImplementedError



# ---
# jupyter:
#   jupytext:
#     formats: ipynb,py:light
#     text_representation:
#       extension: .py
#       format_name: light
#       format_version: '1.3'
#       jupytext_version: 0.8.6
#   kernelspec:
#     display_name: Python 3
#     language: python
#     name: python3
# ---

# +
from besos import eppy_funcs as ef
from besos import sampling
from besos.problem import EPProblem
from besos.evaluator import EvaluatorEP
from besos.parameters import RangeParameter, Parameter, FieldSelector
from besos import config

from sklearn import svm, pipeline, linear_model
from sklearn.preprocessing import StandardScaler
import matplotlib.pyplot as plt
import pandas as pd
import numpy as np
from ipywidgets import interact, FloatSlider
# -

# ## Setup
#
# We begin by loading a representation of a builing. The default is a small office, which will be used for this example.

building = ef.get_building()

# Then define how we would like to modify the building. We choose to change the direction that it is facing, and modify the efficency of the lights.

parameters = [
    Parameter(FieldSelector(object_name='NonRes Fixed Assembly Window',
                                field_name='Solar Heat Gain Coefficient'),
                  value_descriptor=RangeParameter(0.01,0.99)),
    Parameter(FieldSelector('Lights', '*', 'Watts per Zone Floor Area'),
                  value_descriptor=RangeParameter(8, 12),
                  name='Lights Watts/Area')]

# Now we specify that we would like to measure the electricity use of the entire facility. We then bundle all of this information together.

# ## Simulation

# +
objectives = ['Electricity:Facility']

problem = EPProblem(parameters, objectives)
# -

# We can generate some samples train the model on them. If you do not want to wait for the simulation, skip the next code cell.
#
# The sampler uses latin hypercube sampling to distribute the samples, and scales them appropriately.
#
# Here are the inputs that it

inputs = sampling.dist_sampler(sampling.lhs, problem, 5)
inputs

evaluator = EvaluatorEP(problem, building)
train = evaluator.df_apply(inputs, keep_input=True)
train.head() # display the first 5

# We can also load some saved simulation results for this building. 1000 results are loaded.

# +
#train = pd.read_pickle(config.data_dir.joinpath('SHGC_and_lights.p'))

#train.head()
# -

x,y,c = problem.names()
x,y,c

# Here is a plot of the training data we will use.

train.plot.hexbin(x=x, y=y, C=c, cmap='plasma', sharex=False, gridsize=50)

# +
def train_and_plot(density=80):
    model = pipeline.make_pipeline(
        StandardScaler(),
        linear_model.Ridge())
    model.fit(train[[x, y]].values, train[c].values)
    # works for at least density <= 1000, may cause some holes beyond that
    df = get_plot_data(model, density=int(density*1.5))
    df.plot.hexbin(x=x, y=y, C=0, cmap='plasma', gridsize=density, sharex=False)
    return model

def get_plot_data(model, density):
    p1 = problem.inputs[0].value_descriptor
    a = np.linspace(p1.min, p1.max, density)
    p2 = problem.inputs[1].value_descriptor
    b = np.linspace(p2.min, p2.max, density)
    plot_data = pd.DataFrame(np.transpose([np.tile(a, len(b)), np.repeat(b, len(a))]),
                             columns=problem.names('inputs'))
    return pd.concat([plot_data, pd.Series(model.predict(plot_data))], axis=1)
# -

model = train_and_plot()

# +
# this cell uses the model to generate the requested prediction

(min1, max1), (min2, max2) = ((float(p.value_descriptor.min ), p.value_descriptor.max)
                              for p in problem.inputs)
density = 30
df = get_plot_data(model, int(density*1.5))

def model_wrapper(North_Axis, Watts):
    df.plot.hexbin(x=x, y=y, C=0, cmap='plasma', gridsize=30, sharex=False)
    x_lims = plt.xlim()
    y_lims = plt.ylim()
    plt.scatter(x=[North_Axis], y=[Watts])
    plt.xlim((x_lims))
    plt.ylim((y_lims))

    return model.predict([[North_Axis, Watts]])[0]

def avg(a,b):
    return (a+b)/2

continuous_update=False
interact(model_wrapper,
         North_Axis=FloatSlider(min=min1, max=max1, value=avg(min1, max1),
                                step=0.01, continuous_update=continuous_update),
         Watts=FloatSlider(min=min2, max=max2, value=avg(min2, max2),
                           continuous_update=continuous_update))



import pytest

import eppy_funcs as ef
from evaluator import EvaluatorEP
from parameters import RangeParameter, CategoryParameter, Parameter, FieldSelector, expand_plist
from problem import EPProblem, Problem
import platypus
import optimizer
from objectives import MeterReader
import sampling
import random
import numpy as np

#This test takes about 3-4 minutes to run so you can omit it by running pytest_short.py instead of pytest

@pytest.fixture
def parameters():
    parameters = expand_plist(
        {'NonRes Fixed Assembly Window':
        {'UFactor':(0.1,5),
        'Solar Heat Gain Coefficient':(0.01,0.99)
        },
        'Mass NonRes Wall Insulation':{'Thickness':(0.01,0.09)},
        })
    return parameters


@pytest.fixture
def problem(parameters):
    objectives = ['Electricity:Facility', 'Gas:Facility']
    return EPProblem(parameters, objectives)


def test_flexibility(problem):
    """to make sure that you can use multiple algorithms on the same data set"""

    idf = ef.get_idf()
    evaluator = EvaluatorEP(problem, idf)

    random.seed(1)
    #run the first algorithm
    platypus_problem = evaluator.to_platypus()
    algorithm = platypus.NSGAII(problem=platypus_problem, population_size=5)
    algorithm.run(5)

    #run the second algorithm
    generator = platypus.InjectedPopulation(algorithm.population)
    alg2 = platypus.EpsMOEA(problem=platypus_problem, generator=generator, epsilons=3, population_size=5)
    alg2.run(5)

    results = optimizer.solutions_to_df(alg2.result, problem, parts=['inputs', 'outputs'])

    value = results.iloc[0]['Electricity:Facility']
    assert np.isclose(value, 1747893172.6172004), f'Unexpected result: {value}'
    assert all([optimal == True for optimal in results['pareto-optimal']]), f'Algorithm not producing optimal outputs'

    #change this to 0 to see stdout and stderr
    assert 1



import pytest

import eppy_funcs as ef
import sampling
import objectives
from problem import EPProblem
from evaluator import EvaluatorEP, EvaluatorSR
from parameters import RangeParameter, CategoryParameter, expand_plist
import optimizer
import numpy as np

from sklearn import linear_model
from sklearn.model_selection import train_test_split


@pytest.fixture
def building():
    return ef.get_building()


@pytest.fixture
def problem():
    parameters = expand_plist(
    {'Mass NonRes Wall Insulation':
     {'Thickness': (0.01, 0.99)},
     'NonRes Fixed Assembly Window':
     {'U-Factor':(0.1,5),
      'Solar Heat Gain Coefficient':(0.01,0.99)}
    })
    return EPProblem(parameters, ['Electricity:Facility'])


@pytest.fixture
def samples(problem):
    return sampling.dist_sampler(sampling.seeded_sampler, problem, 10)


def test_fit(building, problem, samples):
    """check to make sure linear regression works and is close to the actual value"""

    evaluator = EvaluatorEP(problem, building)
    outputs = evaluator.df_apply(samples)

    train_in, test_in, train_out, test_out = train_test_split(samples, outputs, test_size=0.2)

    reg = linear_model.LinearRegression()
    reg.fit(train_in, train_out)
    results = test_in.copy()
    results['energy use'] = test_out
    results['predicted'] = reg.predict(test_in)

    actual = results.iloc[0]['energy use']
    predicted = results.iloc[0]['predicted']
    assert np.isclose(actual - predicted, 8666133.340038776), f'Unexpected difference of value: {actual - predicted}'

    #change this to 0 to see stdout and stderr
    assert 1


def test_optimisation_with_surrogate(building, problem, samples):
    """make sure we can run optimisation algorithms on the fit surrogates"""

    evaluator = EvaluatorEP(problem, building)
    outputs = evaluator.df_apply(samples)

    train_in, test_in, train_out, test_out = train_test_split(samples, outputs, test_size=0.2)

    reg = linear_model.LinearRegression()
    reg.fit(train_in, train_out)

    def evaluation_func(ind):
        return ((reg.predict([ind])[0][0],),())

    surrogate = EvaluatorSR(evaluation_func, problem)

    s = optimizer.NSGAII(surrogate, 1000)

    optimal = s.iloc[0]['pareto-optimal']
    non_optimal = s.iloc[1]['pareto-optimal']

    assert optimal, f'Optimal ouput was displayed as not Pareto Optimal'
    assert not non_optimal, f'Non-optimal output was displayed as Pareto Optimal'

    #change this to 0 to see stdout and stderr
    assert 1



# ---
# jupyter:
#   jupytext:
#     formats: ipynb,py:light
#     text_representation:
#       extension: .py
#       format_name: light
#       format_version: '1.3'
#       jupytext_version: 0.8.6
#   kernelspec:
#     display_name: Python 3
#     language: python
#     name: python3
# ---

# # Parametric Model
# This notebook covers how to define a parametric eppy model, and get it's energy use for different parameter values.
# ### Various Imports
# `EvaluatorEP` manages energy-plus simulations for a single problem.  
# Problems, like `EPProblem` organise several parameters together.  
# Parameters, like `Parameter` describe a single variable for the building.
# They are composed of a `Descriptor` and a `Selector`.
# `RangeParameter` is a descriptor that indicates a parameter that can take on values from an interval.  
# `CategoryParameter` is a descriptor that indicates a paramter that can take on values from a list.
#

# +
from besos import eppy_funcs as ef
from besos import sampling
from besos.evaluator import EvaluatorEP
from besos.parameters import RangeParameter, CategoryParameter, expand_plist, FieldSelector, Parameter, wwr
from besos.problem import EPProblem

import pandas as pd
# -

# Each building is described by an `.idf` or `epJSON` file. In order to modify it programatically, we load it as a python object.

# by default this will get an example building specifed in the config file.
building = ef.get_building()
# we can specify our own files too.
# (changing the idd file can cause issues, only use one per script/notebook)
# when in idf mode the arguments are building=idf file, data_dict=idd file
# when in json mode the arguments are 

# [Eppy's documentation](https://eppy.readthedocs.io/en/latest/) describes how to explore and modify the idf object. If you are using the newer JSON format, then any JSON parsing library will work.  
# This example uses the fact that the example building contains the following items:
#  - Various `Material` class objects with the 
#    - `Mass NonRes Wall Insulation` (having a `Thickness` field)
#    - `8IN CONCRETE HW` and
#    - `HW CONCRETE` 
#  - A `Construction` class object called `ext-slab` with an `Outside Layer` field
#  - A `WindowMaterial:SimpleGlazingSystem` class object with `UFactor` and `Solar Heat Gain Coefficient` fields
#  - Multiple `Lights` class objects, all of which have a `Watts per Zone Floor Area` field.
#
#  The names class, object and field refer to the 3 layers of nesting present in energyplus building description.

# ## Parameters
# Each way in which the building can be modified is represented by a single parameter. These can be created in various ways, and can be used to automatically modify the idf, as well as to optimise it's design.
# ### Making Parameters
# Parameters are composed of a `Selector` and a `Descriptor`. They can also optionally have a name.  
# There are notebooks which go into more detail on each of these.
# #### Selectors
# Selectors identify which part of the building to modify, and how to modify it. The most common type of selector is a `FieldSelector`. `FieldSelectors` can be created by specifying the class, object, and field names that they apply to.  
#
# The example building loaded above contains a `Material` class object named `Mass NonRes Wall Insulation` which has a `Thickness` field. Below is how to make a selector that modifies this insulation's thickness.

insulation = FieldSelector(class_name='Material', 
                           object_name='Mass NonRes Wall Insulation',
                           field_name='Thickness')

# #### Descriptors
# Descriptors specify what kinds of values are valid for a parameter. If we want to vary a parameter from zero to one excluding the endpoints, we can use a `RangeParameter`, with the apropriate minimum and maximum. (Note that values like 0.001 are also excluded by this example.)

zero_to_one_exclusive = RangeParameter(min_val = 0.01, max_val=0.99)

insulation_param = Parameter(selector=insulation,
                                 value_descriptor=zero_to_one_exclusive,
                                 name='Insulation Thickness')
print(insulation_param)

# Sometimes you might want to specify several `RangeParameter`s at once. The `expand_plist` funcion can do this more concisely. It takes a nested dictionary as input.  
# The keys in the first layer of this dictionary are the names of the idf objects to make parameters for.  
# These are associated with a dictionary that has keys matching the Fields of that object which should be modified. Each field-key corresponds to a tuple containing the minimum and maximum values for that parameter.
# The `class_name` is not specified. Instead the building is searched for objects with the correct `object_name`.

# +
more_parameters = expand_plist(
    # class_name is NOT provided
    #{'object_name':
    # {'field_name':(min, max)}}
    {'NonRes Fixed Assembly Window':
     {'U-Factor':(0.1,5),
      'Solar Heat Gain Coefficient':(0.01,0.99)
     }
    })

for p in more_parameters:
    print(p)
# -

# BESOS also includes some premade parameters, such as window to wall ratio. These can be customised, but for now we will just use the defaults.

# use a special shortcut to get the window-to-wall parameter
window_to_wall = wwr()
print(window_to_wall)

# Parameters can also be created by defining a function that takes an idf and a value and mutates the idf. These functions can be specific to a certain idf's format, and can perform any arbitrary transformation. Creating these can be more involved, and is not covered in this example.

# ## Problems
# problem objects represent the inputs and outputs that we are interested in. We have defined the inputs using parameters above, and will use the default output of electricity use, and the default of no constraints.

parameters = [insulation_param] + more_parameters + [window_to_wall]
problem = EPProblem(inputs=parameters)

# ## Sampling
# Once you have defined your parameters, you may want to generate some random possible buildings. Sampling functions allow you to do this.

samples = sampling.dist_sampler(sampling.lhs, problem, num_samples=3, criterion='maximin')
# arguments can be passed to the specific sampler, here criterion is optional
samples

# ## Evaluation
# Now we can evaluate the samples. We create an energy plus evaluator (`EvaluatroEP`) using the parameters, and idf describing the building, and the objectives we want to measure. For this example we will just use one of the premade objectives: Electricity use for the whole facility.

evaluator = EvaluatorEP(problem, building)
# The evaluator will output electricity use by default
outputs = evaluator.df_apply(samples ,keep_input=True)
# outputs is a pandas dataframe with one new column since only one objective was requested

outputs





import pytest

from parameters import CategoryParameter, RangeParameter, Parameter
from evaluator import EvaluatorSR
from problem import Problem
import optimizer
import numpy as np

from platypus.config import PlatypusConfig
import random

@pytest.fixture
def parameters():
    parameters=[Parameter(value_descriptor=CategoryParameter(list(range(80,90))), name='1st'),
            Parameter(value_descriptor=CategoryParameter(list(range(80,90))), name='2nd'),
            Parameter(value_descriptor=RangeParameter(3, 100), name='3rd')
           ]
    return parameters


def mixed_types(vals):
    #helper function from the jupyter notebook
    vals = list(vals)
    num = vals.pop(-1)
    objectives = tuple(num % v for v in vals)
    return (objectives,())


def test_mixed_type_parameters(parameters):
    """to make sure that we can use different parameter types in the same algorithms while still getting optimal outputs"""

    evaluator = EvaluatorSR(mixed_types, Problem(parameters, 2))
    random.seed(1)
    results = optimizer.EpsMOEA(evaluator, epsilons=10)

    value = results.iloc[0]['outputs_0']
    print(value)
    assert np.isclose(value, 0.4451952793850751), f'Unexpected output: {value}'

    assert 1

def test_optimizer_customization(parameters):
    """testing the variator field while using mixed parameters"""

    evaluator = EvaluatorSR(mixed_types, Problem(parameters, 2))
    variator = optimizer.get_operator(evaluator.to_platypus())

    random.seed(2)
    results = optimizer.EpsMOEA(evaluator, variator=variator, epsilons=10)

    value = results.iloc[0]['outputs_1']
    print(value)
    assert np.isclose(value, 0.4607928200940279), f'Unexpected output: {value}'

    assert 1



# ---
# jupyter:
#   jupytext:
#     formats: ipynb,py:light
#     text_representation:
#       extension: .py
#       format_name: light
#       format_version: '1.3'
#       jupytext_version: 0.8.6
#   kernelspec:
#     display_name: Python 3
#     language: python
#     name: python3
# ---

# +
from besos import eppy_funcs as ef
from besos import optimizer
from besos import sampling
from besos.evaluator import EvaluatorEP, EvaluatorSR
from besos.parameters import RangeParameter, expand_plist, wwr
from besos.problem import EPProblem

import pandas as pd
import numpy as np

from sklearn import linear_model
from sklearn.model_selection import train_test_split

import random
# -

idf = ef.get_idf()

# +
parameters = expand_plist(
    #{'Name of object in idf':
    # {'Property Name':(min, max)}}
    {'NonRes Fixed Assembly Window':
     {'UFactor':(0.1,5),
      'Solar Heat Gain Coefficient':(0.01,0.99)
     },
     'Mass NonRes Wall Insulation':{'Thickness':(0.01,0.09)},
    })
parameters.append(wwr())

problem = EPProblem(parameters)
evaluator = EvaluatorEP(problem, idf)

# +
sampler = sampling.lhs
numSamples = 5

inputs = sampling.dist_sampler(sampling.lhs, problem, numSamples)

outputs = evaluator.df_apply(inputs)
# -

train_in, test_in, train_out, test_out = train_test_split(inputs, outputs, test_size=0.2)

reg = linear_model.LinearRegression()
reg.fit(train_in, train_out)
#r2 score

# +
def evaluation_func(ind):
    return ((reg.predict([ind])[0][0],),())

surrogate = EvaluatorSR(evaluation_func, problem)
# -

surrogate.df_apply(test_in, keep_input=True)

predictions = test_in.copy()
predictions['prediction'] = surrogate.df_apply(predictions)#reg.predict(test_in)
predictions['answer'] = test_out
predictions

s = optimizer.NSGAII(surrogate, 1000)

s





from typing import Union, List, Dict, Callable, Iterable

import pandas as pd
import numpy as np
import platypus
import warnings

import IO_Objects
import config
import objectives
import parameters

# from pyehub_funcs import name_getter


# TODO: Consider storing the constraint bounds with the constraints themselves, not in the problem
# also consider storing the direction of optimisation inside the objectives
# might be able to inherit some of the constraint parsing from platypus, not sure if that is worth the hassle
class Problem(IO_Objects.ReprMixin):
    """A class that collects all of the inputs, outputs and constraints related to a building."""
    valid_parts: List[str] = ['inputs', 'outputs', 'constraints', 'violation']
    default_converters = {'outputs': IO_Objects.Objective, 'constraints': IO_Objects.Objective}

    def __init__(self, inputs: Union[int, List[Union[str, IO_Objects.Descriptor]]] = None,
                 outputs: Union[int, List[Union[str, IO_Objects.Objective]]] = None,
                 constraints: Union[int, List[Union[str, IO_Objects.Objective]]] = None, *,
                 constraint_bounds: List[str] = None, minimize_outputs: List[bool] = None,
                 converters: Dict[str, Callable[[str], IO_Objects.IOBase]] = None):
        """

        :param inputs: A list of Parameters, or an integer.
            If a list  is used, strings are converted to Parameters and this list determines the valid inputs.
            If an integer, this problem accepts that many inputs.
        :param outputs: A list of Objectives, or an integer.
            If a list  is used, strings are converted to Objectives and this list determines the valid inputs.
            If an integer, this problem requires that many outputs
        :param constraints:
        :param constraint_bounds:
        :param minimize_outputs:
        :param converters:
        """
        super().__init__()
        self.converters = converters or self.default_converters
        extra_keys = set(self.converters.keys()) - set(self.valid_parts)
        if extra_keys:
            raise ValueError(f'The keys {extra_keys} are not valid for this Problem. Only {self.valid_parts} are valid')

        self.inputs = self._io_to_list(inputs, 'inputs')
        self.num_inputs = len(self.inputs)

        self.outputs = self._io_to_list(outputs, 'outputs')
        self.num_outputs = len(self.outputs)
        self.minimize_outputs = minimize_outputs or [True] * self.num_outputs
        msg = 'outputs and minimize_outputs must have the same length'
        assert len(self.minimize_outputs) == self.num_outputs, msg

        self.constraints = self._io_to_list(constraints, 'constraints')
        self.num_constraints = len(self.constraints)
        # TODO: consider using platypus's constraints here
        self.constraint_bounds = constraint_bounds or []
        msg = 'constraints and constraint_bounds must have the same length'
        assert len(self.constraint_bounds) == self.num_constraints, msg
        self.fix_names()
        self._add_reprs(['inputs', 'outputs', 'minimize_outputs',
                         'constraints', 'constraint_bounds', 'converters'],
                        check=True)

    def fix_names(self):
        mapping = {}
        duplicates = []
        for obj in self:
            mapping[obj.name] = mapping.get(obj.name, []) + [obj]
        for name, objects in mapping.items():
            if len(objects) != 1:
                duplicates.append((name, objects))
        if duplicates:
            warnings.warn(RuntimeWarning(f'Duplicate names found. (duplicate, repetitions): '
                                         f'{[(name, len(objects)) for name, objects in duplicates]}'
                                         f'\nAttempting to fix automatically'))
        for name, objects in duplicates:
            for i, obj in enumerate(objects):
                obj.name = f'{obj.name}_{i}'

    def _io_to_list(self, io_objects: Union[int, List[IO_Objects.IOBase], None], part):
        """Converts a list of objects to a standard form:
        numbered placeholders, original datatype or io_object that match the part provided.
        """
        if io_objects is None:
            return []
        if isinstance(io_objects, int):
            if part == 'inputs':
                class_ = parameters.Parameter
            elif part in ['outputs', 'constraints']:
                class_ = IO_Objects.Objective
            else:
                raise ValueError(f'Cannot produce dummy values for part {part}')
            return [class_(name=f'{part}_{i}') for i in range(io_objects)]
        if isinstance(io_objects, (str, IO_Objects.IOBase)):
            io_objects = [io_objects]
        return [self.convert(o, part) for o in io_objects]

    def convert(self, io_object, part) -> IO_Objects.IOBase:
        """

        :param io_object: An object that should be converted to a parameter, objective or constraint
        :param part: one of 'inputs', 'outputs' or 'constraints' describing what to convert `io_object` to
        :return: the converted object
        """
        if isinstance(io_object, IO_Objects.IOBase):
            return io_object
        if part in self.converters:
            f = self.converters[part]
            try:
                return f(io_object)
            except TypeError as e:
                try:
                    if isinstance(io_object, dict):
                        return f(**io_object)
                    if isinstance(io_object, Iterable):
                        return f(*io_object)
                except:
                    pass
                raise TypeError(f'Cannot convert {io_object} to {part}') from e
        return io_object

    def expand_parts(self, parts: Union[str, List[str]]) -> List[str]:
        """Expands 'auto' and 'all' to the correct lists of parts, and wraps single parts in a list"""
        if parts == 'auto':
            if self.num_constraints == 0:
                parts = ['inputs', 'outputs']
            else:
                parts = 'all'
        if parts == 'all':
            parts = self.valid_parts
        elif isinstance(parts, str):
            parts = [parts]

        if not set(parts) <= set(self.valid_parts):
            raise ValueError(f"parts must be a subset of {self.valid_parts + ['all']}, not {parts}")
        return parts

    def names(self, parts: Union[str, List[str]] = 'auto') -> List[str]:
        """

        :param parts: one of {'inputs', 'outputs', 'constraints', 'violation', 'all', 'auto'}
        :return: the names requested
        """
        parts = self.expand_parts(parts)
        names = []
        for attr in parts:
            if attr == 'violation':
                names.append('violation')
            else:
                part = getattr(self, attr)
                if part is None:
                    raise ValueError(f'{attr} names not available')
                names.extend(IO_Objects.get_name(i) for i in part)
        return names

    # TODO: Add support for pareto-optimal column
    # TODO: Consolidate the different to_df code (ie from optimizer.py)
    def to_df(self, table: Union[np.array, pd.DataFrame], parts: Union[str, List[str]] = 'auto') -> pd.DataFrame:
        """Converts the given table to a DataFrame that matches this problem's input/output format

        :param table: a table to be converted to a DataFrame. Must have the right number of columns.
        :param parts: inputs, outputs, constraints or all, depending on which data the DataFrame contains
        :return: A DataFrame containing the same data as the original table.
        """
        columns = self.names(parts)
        types = [p.pd_type if hasattr(p, 'pd_type') else None for p in self.expand_parts(parts)]
        if isinstance(table, pd.DataFrame):
            if len(table.columns) != len(columns):
                raise ValueError(f'columns: {columns} requested but {list(table.columns)} found')
            return table[columns]

        df = pd.DataFrame(table, columns=columns)
        # TODO: Make the categorical columns have the type category instead of object (attempt commented out below)
        # for col, type_ in zip(df, types):
        #     if type_:
        #         df[col] = df[col].astype(type_)
        return df

    def partial_df(self, table: Union[np.array, pd.DataFrame], parts='all'):
        parts = self.expand_parts(parts)
        for i in range(1, len(parts) + 1):
            partial_parts = parts[:i]
            try:
                return self.to_df(table, partial_parts), partial_parts
            except ValueError:
                continue
        raise ValueError('Could not find a matching DataFrame')

    def to_platypus(self) -> platypus.Problem:
        """Converts this problem to a platypus problem.
        No evaluator will be included.

        :return: A corresponding platypus problem
        """
        problem = platypus.Problem(self.num_inputs, self.num_outputs, self.num_constraints)
        for i, parameter in enumerate(self.inputs):
            problem.types[i] = parameter.platypus_type
        for i, direction in enumerate(self.minimize_outputs):
            problem.directions[i] = platypus.Problem.MINIMIZE if direction else platypus.Problem.MAXIMIZE
        for i, bound in enumerate(self.constraint_bounds):
            problem.constraints[i] = bound
        return problem

    def __eq__(self, other):
        return (self.__class__ is other.__class__
                and self.inputs == other.inputs
                and self.outputs == other.outputs
                and self.constraints == other.constraints)

    def __iter__(self):
        return iter(self.inputs + self.outputs + self.constraints)


# TODO: consider having shortcuts for the converters instead of making this a whole different class
class EPProblem(Problem):
    """A problem with defaults that are appropriate for EnergyPlus simulations"""
    default_converters = {'outputs': objectives.MeterReader, 'constraints': objectives.MeterReader}

    def __init__(self, inputs=None, outputs=config.objectives,
                 constraints=None, converters=None, **kwargs):
        super().__init__(inputs=inputs, outputs=outputs, constraints=constraints, converters=converters, **kwargs)

class EHProblem(Problem):
    """A problem that works with PyEHub models"""
    #TODO: Restructure if possible to be more like other problems.
    def __init__(self, inputs = [], outputs = ['total_cost'], constraints = None, converters = None, **kwargs):
        super().__init__(inputs=inputs, outputs=outputs, constraints=constraints, converters=converters, **kwargs)

    #Overwritten functions to work with EvaluatorEH:
    def fix_names(self):
        pass
    



# ---
# jupyter:
#   jupytext:
#     formats: ipynb,py:light
#     text_representation:
#       extension: .py
#       format_name: light
#       format_version: '1.3'
#       jupytext_version: 0.8.6
#   kernelspec:
#     display_name: Python 3
#     language: python
#     name: python3
# ---

# This notebook implements a rbf adaptive surrogate. It could be used as the model half of an rbf-opt algorithm, but would need a sample selection algorithm as well.

# +
from besos import eppy_funcs as ef
from besos import sampling
from besos.evaluator import EvaluatorEP, AdaptiveSR, EvaluatorSR
from besos.parameters import RangeParameter, expand_plist, wwr
from besos.problem import EPProblem

import numpy as np
import pandas as pd
import scipy.interpolate

# +
from typing import Union, List, Tuple

tabular = Union[np.array, pd.DataFrame]
DF = pd.DataFrame

class RBF(AdaptiveSR):
    def __init__(self, *args, **kwargs):
        self.model_kwargs = {}
        for name in ['function', 'epsilon', 'smooth', 'norm']:
            if name in kwargs:
                self.model_kwargs[name] = kwargs.pop(name)
        super().__init__(*args, **kwargs)
        self.obj_models = None
        self.constr_models = None

    def train(self) -> None:
        """Generates a new model using the stored data, and stores it as self.model"""
        inputs = (self.data[self.problem.names('inputs')].values[:,i] for i in range(problem.num_inputs))
        outputs = self.data[self.problem.names('outputs')].values
        constraints = self.data[self.problem.names('outputs')].values
        def train_single(output):
            return scipy.interpolate.Rbf(*inputs, output, **self.model_kwargs)
    
        self.obj_models = [train_single(outputs[:,i]) for i in range(self.problem.num_outputs)]
        self.constr_models = [train_single(outputs[:,i]) for i in range(self.problem.num_constraints)]

    def eval_single(self, values: List) -> Tuple:
        values = ([v] for v in values)
        objectives = tuple(model(*values)[0] for model in self.obj_models)
        constraints = tuple(model(*values)[0] for model in self.constr_models)
        return (objectives, constraints)
    
    def do_infill(self, data: DF) -> None:
        """Updates the model using the inputs X and outputs y, and stores the added data

        :param data: a table of training data
        :return: None
        """
        old_df = self.data
        df, parts = self.problem.partial_df(data, parts=['inputs', 'outputs', 'constraints'])
        if parts == ['inputs']:
            outputs: DF = self.get_from_reference(df)
            df = pd.concat([df, outputs], axis=1)
        self.append_data(df)
        self.train()

# +
building = ef.get_building()

parameters=expand_plist(
    {'Mass NonRes Wall Insulation':
      {'Thickness': (0.01, 0.99)},
     'NonRes Fixed Assembly Window':
     {'U-Factor':(0.1,5),
      'Solar Heat Gain Coefficient':(0.01,0.99)}
    })

parameters.append(wwr())

problem = EPProblem(parameters)

# a fast function for debugging
def placeholder(values):
    return (sum(values),),()

evaluator = EvaluatorEP(problem, building)
#evaluator = EvaluatorSR(placeholder, problem)

inputs = sampling.dist_sampler(sampling.lhs, problem, 10)
# -

k = RBF(reference=evaluator)

k.do_infill(inputs)

k.df_apply(inputs, keep_input=True)





# ---
# jupyter:
#   jupytext:
#     formats: ipynb,py:light
#     text_representation:
#       extension: .py
#       format_name: light
#       format_version: '1.3'
#       jupytext_version: 0.8.6
#   kernelspec:
#     display_name: Python 3
#     language: python
#     name: python3
# ---

# +
from besos import eppy_funcs as ef
from besos import sampling
from besos.evaluator import EvaluatorEP
from besos.parameters import FieldSelector, FilterSelector, GenericSelector, Parameter
from besos.problem import EPProblem

import pandas as pd
# -

building = ef.get_building(mode='json')
# we need the json example file because of how the insulation_filter function works

# Selectors identify which part of the building to modify, and how to modify it.
# `FieldSelector`s allow for modifying individual fields in a building.  
#
# The example building loaded above contains a `Material` class object named `Mass NonRes Wall Insulation` which has a `Thickness` field. Below is how to make a selector that modifies this insulation's thickness.

FieldSelector(class_name='Material', object_name='Mass NonRes Wall Insulation', field_name='Thickness')

# However, there is only one object in the example building with the name `Mass NonRes Wall Insulation`, so we can ommit the `class_name`. The building will be searched for any object with the correct `object_name`.

FieldSelector(object_name='Mass NonRes Wall Insulation', field_name='Thickness')

#  If the `class_name` is provided, but the `object_name` is omitted, then the first object with that `class_name` will be used. Since JSON files do not guarentee ordering, this only works for idf files. `field_name` is mandatory.

# `FilterSelectors` allow us to use custom function to select the objects to modify. Here we define a function that finds all materials with `Insulation` in their name. Then we use this function to modify the thickness of all materials in the building.

# +
def insulation_filter(building):
    return [obj for name, obj in building['Material'].items() if 'Insulation' in name]

insulation = FilterSelector(insulation_filter, 'Thickness')
insulation
# -

# When you have multiple objects of the same type that all share the same field, and want to vary that field's value, you can set the `object_name` to `'*'`, and the selector will modify all of those objects at once.

lights = FieldSelector(class_name='Lights', object_name='*', field_name='Watts per Zone Floor Area')
lights

# Parameters can also be created by defining a function that takes an idf and a value and mutates the idf. These functions can be specific to a certain idf's format, and can perform any arbitrary transformation. Creating these can be more involved.  
# `eppy_funcs` contains the functions `one_window` and `wwr_all`. `one_window` removes windows from a building untill it has only one per wall. `wwr_all` takes a building with one window per wall and adjusts it to have a specific window to wall ratio.

window_to_wall = GenericSelector(set=ef.wwr_all, setup=ef.one_window)
window_to_wall

# ### Sampling
# Since Selectors do not describe the values they can take on, only where those values go, they are not sufficient for sampling. We can specify our samples manually.

samples = pd.DataFrame({
    'Thickness': [x/10 for x in range(1,10)]*2,
    'Watts': [8,10,12]*6,
    'wwr': [0.25, 0.5]*9
})
samples.head()

# We also need to put the selectors in parameters before we can use them in an evaluator.

# +
parameters= [Parameter(selector=x) for x in(insulation, lights, window_to_wall)]
problem = EPProblem(inputs=parameters)

evaluator = EvaluatorEP(problem, building)
# The evaluator will use this objective by default
outputs = evaluator.df_apply(samples ,keep_input=True)
# outputs is a pandas dataframe with one column since only one objective was requested
# -

outputs





# ---
# jupyter:
#   jupytext:
#     formats: ipynb,py:light
#     text_representation:
#       extension: .py
#       format_name: light
#       format_version: '1.3'
#       jupytext_version: 0.8.6
#   kernelspec:
#     display_name: Python 3
#     language: python
#     name: python3
# ---

# +
from eplus_ga import eppy_funcs as ef
from eplus_ga import sampling
from eplus_ga import objectives

from eplus_ga.problem import EPProblem
from eplus_ga.evaluator import EvaluatorEP

from sklearn import linear_model
from sklearn.model_selection import train_test_split

from eplus_ga.parameters import RangeParameter, CategoryParameter, expand_plist
# -

building = ef.get_building()

# +
parameters = expand_plist(
    {'Mass NonRes Wall Insulation':
     {'Thickness': (0.01, 0.99)},
     'NonRes Fixed Assembly Window':
     {'U-Factor':(0.1,5),
      'Solar Heat Gain Coefficient':(0.01,0.99)}
    })

# parameters.append(RangeParameter.wwr())

problem = EPProblem(parameters, ['Electricity:Facility'])

# +
#Adding categorical parameters requires converting them for scikit-learn
# -

inputs = sampling.dist_sampler(sampling.lhs, problem, 10)

evaluator = EvaluatorEP(problem, building)
outputs = evaluator.df_apply(inputs)

train_in, test_in, train_out, test_out = train_test_split(inputs, outputs, test_size=0.2)

reg = linear_model.LinearRegression()
reg.fit(train_in, train_out)
results = test_in.copy()
results['energy use'] = test_out
results['predicted'] = reg.predict(test_in)
results

inputs





"""
This module provides helper functions for interacting with PyEHub.
"""
import contextlib
import config
from functools import partial, reduce
import numpy as np
import operator
import pandas as pd
import warnings

from errors import ModeError
import pyehub 
from pyehub.energy_hub.ehub_model import EHubModel

"""This file is a collection of functions to interact with PyEHub."""

#TODO: replace back with yaml loader?
def get_xlsx(excel_file: str = config.files.get("xlsx")):
    return excel_file

def get_hub():
    """Generates the base PyEHub model from the excel file."""
    # could check the settings to see what type of ehub model to create
    data_file = get_xlsx()
    model = pyehub.energy_hub.EHubModel(excel=data_file)
    return model

def get_by_path(root, items):
    """Access a nested object in root by item sequence. Used to navigate PyEHub's not flat dict."""
    #TODO: inset try statement here incase the parameter doesn't exist to fail gracefuly
    return reduce(operator.getitem, items, root)

def set_by_path(root, items, value):
    """Set a value in a nested object in root by item sequence. Used to set values in PyEHub's not flat dict."""
    get_by_path(root, items[:-1])[items[-1]] = value
    
# no longer necessary due to restructure of EHProblem    
def name_getter(dict):
    """Get list of names from a dictionary. Used to get the objective names for EHProblem."""
    namelist = []
    for item in dict:
        name = ''
        for part in item:
            if(name == ''):
                name = name + part
            else:
                name = name + ':' + part  
        namelist.append(name)
    return namelist

def pyehub_parameter_editor(hub, parameters, values:list):
    """Changes the __dict__ of the energy hub for the parameters specified
    on initialization with the values given to evaluator."""
    if (len(parameters) > 0): # There are paremeters to edit
        for (parameter,value) in zip(parameters,values):
            # Change the value of the paremeter in the model's dict
            set_by_path(hub.__dict__,parameter,value)


def _split_EP_days(hourly_series):
    """
    Splits the hourly series of EP simulations that simulate multiple days in different seasons into specific energyhub time series.
    """
    #TODO: replace with a better loop of checking for the number of hours in the series and splitting better
    result1 = hourly_series.head(24)
    result1 = result1.reset_index()
    result2 = hourly_series.tail(24)
    result2 = result2.reset_index()
    return result1, result2

    #TODO: 



def ep_to_eh(input, index_size):
    """
    Converts EnergyPlus evaluator outputs to PyEHub compatible inputs.
    """
    #TODO: Needs to be finished

    input_dict = input.to_dict()

    # Need to have a better joining on index for the pandas dataframe instead of appending (difficulties to be noted with inserting dictionaries into cells)
    columnnames = []
    for j in input_dict:
        columnnames.append(j)

    #2x for the number of seasons in the splitting will need to have some value stored from the smarter splitting 
    # or make an easy spot to modify a value that propigates through the whole system.
    #if using the input for the function
    #index_size *=2
    # which should be something that does this
    # index_size = len(EPdf)*2
    # df_input = pd.DataFrame(np.nan, index=range(0,index_size), columns=columnnames)
    df_input = pd.DataFrame(columns=columnnames)

    for objective in input_dict:
        for series in input_dict[objective]:
            result = input_dict[objective][series].to_frame()
            #TODO: make a more genaric splitting process
            cold_result, warm_result = _split_EP_days(result)

            #TODO: make a more generic unit conversion function 
            #3.6 Megajoules per kWh
            cold_result = cold_result /3600000
            warm_result = warm_result / 3600000
            cold_dict = cold_result.to_dict()
            cold_dict = cold_dict['Value']
            warm_dict = warm_result.to_dict()
            warm_dict = warm_dict['Value']

            cold_input = [cold_dict]
            warm_input = [warm_dict]

            temp_df1 = pd.DataFrame(np.array(cold_input),columns = [j])
            temp_df2 = pd.DataFrame(np.array(warm_input),columns = [j])
            #instead of appending I can edit the value of the specific cell with the input (ah that gets difficult with dictionaries) 
            # df_input.at[i, j] = [cold_dict]
            # df_input.iloc[i+len(EPdf), df_input.columns.get_loc(j)] = warm_dict
            # could do some sort of join on the 
            df_input = df_input.append(temp_df1, ignore_index = True)
            df_input = df_input.append(temp_df2, ignore_index = True)

    return df_input


import pytest

from optimizer import rbf_opt
from evaluator import EvaluatorSR
from parameters import Parameter, RangeParameter
from problem import Problem
import numpy as np

def test_rbfopt():
    """Simple test to make sure RBFopt can be run properly"""

    def obj_funct(x):
        return (x[0]*x[1] - x[2],),()

    param_list = [Parameter(value_descriptor=RangeParameter(0,10)) for _ in range(3)]
    problem = Problem(param_list, 1)

    evaluator = EvaluatorSR(obj_funct, problem, error_mode='Silent')

    opt = rbf_opt(evaluator, 30)
    value = opt.iloc[0]

    assert np.isclose(value['None_0'], 0) and (np.isclose(value['None_1'], 5.062882681226724) or np.isclose(value['None_1'], 0.5148690)) and np.isclose(value['None_2'], 10) and np.isclose(value['outputs_0'], -10), f'Unexpected output: {value}'     # None_1 assert values hardcoded to pass on Debian



# ---
# jupyter:
#   jupytext:
#     formats: ipynb,py:light
#     text_representation:
#       extension: .py
#       format_name: light
#       format_version: '1.3'
#       jupytext_version: 0.8.6
#   kernelspec:
#     display_name: Python 3
#     language: python
#     name: python3
# ---

# +
from besos import eppy_funcs as ef

from besos.parameters import RangeParameter, expand_plist, wwr
from besos.evaluator import EvaluatorEP
from besos.problem import EPProblem
from besos import optimizer
import platypus

# +
building = ef.get_building()

parameters=expand_plist(
    {'NonRes Fixed Assembly Window':
     {'U-Factor':(0.1,5),
      'Solar Heat Gain Coefficient':(0.01,0.99)
     },
     'Mass NonRes Wall Insulation':{'Thickness':(0.01,0.09)},
    })
parameters.append(wwr())

objectives = ['Electricity:Facility', 'Gas:Facility']

problem = EPProblem(parameters, objectives)

evaluator = EvaluatorEP(problem, building)
# -

s = optimizer.NSGAII(evaluator, evaluations=10, population_size=2)
s

# +
# s = optimizer.OMOPSO(evaluator, evaluations=30, population_size=10, epsilons=4)
# s = Optimizer.EpsMOEA(evaluator, population_size=2, evaluations=6)
# s = Optimizer.GDE3(evaluator, population_size=10, evaluations=20)
# s = Optimizer.SPEA2(evaluator, population_size=10, evaluations=20)
# s = Optimizer.NSGAIII(evaluator,population_size=10, evaluations=20)
# -





import os
import hublib.ui as ui
import ipywidgets as w
from IPython.display import display
import config

import eppy_funcs as ef
import sampling

from functools import partial


class FileManager:
    def __init__(self, extensions, dir=config.inDir, defaults=config.files, **kwargs):
        self.files = {}  # maps extension name to (current, {candidates}) tuples
        # defaults must already be in the directory
        for ext, file in defaults.items():
            if ext not in extensions:
                raise ValueError('{} is not a valid extension'.format(ext))
            self.files[ext] = (file, {file})
        self.directory = dir
        self.extensions = extensions
        self.out = w.Output()
        self.displayOptions()

        widgetDefaults = {'name': "Files", 'desc': '',
                          'maxsize': '10M', 'width': '30%', 'maxnum': 5}
        widgetDefaults.update(kwargs)

        multiLoader = ui.FileUpload(dir=dir, cb=self, **widgetDefaults)
        self.w = w.VBox([multiLoader.w, self.out])

    def __call__(self, w, fileNames):
        self.out.clear_output()
        badFiles = []
        with self.out:
            for name in fileNames:
                ext = name.split('.')[-1]
                name = self._short(name)
                if ext in self.extensions:
                    current, options = self.files.get(ext, ('File Needed', set()))
                    self.files[ext] = (name, options | {name})
                else:
                    badFiles.append(name)
            if badFiles:
                print(' and '.join([', '.join(badFiles[:-1]), badFiles[-1]]),
                      'are invalid files. Only files with the extensions',
                      ' or '.join([', '.join(self.extensions[:-1]), self.extensions[-1]]),
                      'are allowed')
        self.displayOptions()
        w.reset()

    def displayOptions(self):
        with self.out:
            for ext in self.extensions:
                current, options = self.files.get(ext, ('File Needed', ['File Needed']))
                dd = ui.Dropdown(name=ext + ' file', description='',
                                 value=current,
                                 options=options, cb=partial(self._setFile, ext=ext))
                display(dd)

    def _short(self, name):
        return name[len(self.directory) + 1:]

    def getFile(self, ext):
        # deliberetly does not catch key errors
        # might want to replace with __getitem__
        return os.path.join(self.directory, self.files[ext][0])

    def _setFile(self, w, file, ext):
        current, options = self.files[ext]
        self.files[ext] = (file, options)


fileTypes = ['building', 'idd', 'epw']
fm = FileManager(fileTypes)


def output(transformers):
    idf = None
    results = w.Output()

    def _getSamples(w):
        nonlocal idf
        global df
        if idf is None:
            idf = ef.get_idf(fm.getFile('building'), fm.getFile('idd'), fm.getFile('epw'))
        with results:
            print('Simulating...')
            df = sampling.dist_sampler(samplerType.value, transformers, samplesBox.value)
            df['energy use'] = ef.Evaluator(transformers, idf).df_apply(df)
            print('Done')

            display(df)
    goButton = w.Button(description='Get Samples')
    goButton.on_click(_getSamples)
    return w.VBox([goButton, results])


startButtons = w.HBox([ui.HideCodeButton().w, ui.RunAllButton().w])

samplerType = ui.Dropdown('Sampling method', '',
                          options={'Latin Hypercube': sampling.lhs,
                                   'Random': sampling.random,
                                   'LHS with minimax criterion': sampling.lhs_maximin,
                                   'Random, seeded': sampling.seeded_sampler},
                          value=sampling.lhs)

samplesBox = ui.Integer('Samples', value=5, min=1, max=20)

interface = w.VBox([fm.w, samplerType.w, samplesBox.w])



import pytest

import eppy_funcs as ef
from evaluator import EvaluatorEP, EvaluatorSR
from parameters import RangeParameter, CategoryParameter, Parameter, FieldSelector, wwr
from problem import EPProblem, Problem
import sampling
import numpy as np

#parameter - selector & descriptor
#selector - FilterSelector FieldSelector GenericSelector
#descriptor - RangeParameter CategoryParameter

#init, setup_checks, and setup_changes were moved directly from Will's code in parameters.py
def test_init():
    # inputs should be initialisable
    Parameter(FieldSelector(object_name='NonRes Fixed Assembly Window', field_name='UFactor'),
              value_descriptor=RangeParameter(min_val=0.1, max_val=5))
    wwr()
    wwr(name='other')
    print('Parameters initialised')

    #change this to 0 to see stdout and stderr
    assert 1


#TODO: look into fixing this
def test_setup_checks():
    #try excepts were updated to asserts and now the test fails, no exceptions are being raised and it seems to be in 3rd party code

    idf = ef.get_idf()
    assert len(idf.idfobjects['FENESTRATIONSURFACE:DETAILED']) == 21, 'bad idf'

    # eppy key-inputs should reject bad key values at setup
    r: Parameter = Parameter(FieldSelector(class_name='invalid', field_name='any'))

    #when exception raising is implemented properly uncomment this test, for now it does nothing

    #with pytest.raises(Exception):
    #    r.setup(idf)
    #print('Invalid object key detected successfully.')
    #r = Parameter(FieldSelector(object_name='NonRes Fixed Assembly Window', field_name='invalid'))
    #with pytest.raises(Exception):
    #    r.setup(idf)
    #print('Invalid property detected successfully.')

    #change this to 0 to see stdout and stderr
    assert 1


def test_setup_changes():
    idf = ef.get_idf()

    r1 = Parameter(FieldSelector(object_name='NonRes Fixed Assembly Window', field_name='UFactor'),
                   value_descriptor=RangeParameter(min_val=0.1, max_val=5))
    r2 = wwr()
    r3 = wwr(name='other')

    r1.setup(idf)
    r2.setup(idf)
    assert len(idf.idfobjects['FENESTRATIONSURFACE:DETAILED']) == 4, 'bug in r2.setup'

    idf = ef.get_idf()
    assert len(idf.idfobjects['FENESTRATIONSURFACE:DETAILED']) == 21, 'bad idf'
    r3.setup(idf)
    assert len(idf.idfobjects['FENESTRATIONSURFACE:DETAILED']) == 4, 'bug in r3.setup'

    print('setup tests done')

    #change this to 0 to see stdout and stderr
    assert 1


def test_custom_evaluation():
    """check to see if descriptors display as intended, and check to make sure custom evaluations work with EvaluatorSR"""

    #create the descriptors
    zero_to_nine = RangeParameter(min_val = 0, max_val=9)
    single_digit_integers = CategoryParameter(options=[0,1,2,3,4,5,6,7,8,9])
    text_example = CategoryParameter(options=['a', 'b', 'c', 'other'])

    #create the parameters and the problem
    parameters = [
        Parameter(value_descriptor=zero_to_nine, name='0-9'),
        Parameter(value_descriptor=single_digit_integers, name='single digit'),
        Parameter(value_descriptor=text_example, name='text')
    ]
    problem = Problem(parameters, outputs=['output'])

    #create the sampling distribution (seeded with only one output that should give     4.939321535345923	7	c	34.575251 after evaluating)
    samples = sampling.dist_sampler(sampling.seeded_sampler, problem, num_samples=1)
    values = [samples.iloc[0]['0-9'], samples.iloc[0]['single digit'], samples.iloc[0]['text']]
    assert np.isclose(values[0], 4.939321535345923) and values[1:] == [7, 'c'], f'Unexpected sample values: {values}'

    #custom evaluation function from the jupyter notebook
    def evaluation_function(values):
        x,y,z = values
        if z == 'other':
            return (0,), ()
        else:
            return (x * y,), ()

    evaluator = EvaluatorSR(evaluation_function, problem)
    # The evaluator will use this objective by default
    outputs = evaluator.df_apply(samples ,keep_input=True)
    result = outputs.iloc[0]['output']
    print(result)
    assert np.isclose(result, 34.57525074742146), f'Unexpected evaluation result: {result}'

    #change this to 0 to see stdout and stderr
    assert 1



import pytest

import eppy_funcs as ef
from evaluator import EvaluatorEP, EvaluatorSR, EvaluatorEH
import numpy as np
import pandas as pd
from parameters import RangeParameter, CategoryParameter, Parameter, FieldSelector
from problem import EPProblem, Problem, EHProblem
import pyehub_funcs as pf
import sampling

@pytest.fixture
def building():
    #returns the basic building
    return ef.get_building()

@pytest.fixture
def hub():
    #returns the basic hub
    return pf.get_hub()

@pytest.fixture
def problem():
    parameters = [Parameter(FieldSelector(object_name='Mass NonRes Wall Insulation', field_name='Thickness'))]
    objectives = ['Electricity:Facility', 'Gas:Facility'] # the default is just 'Electricity:Facility'

    problem=EPProblem(parameters, objectives) #EPP Problem automatically converts these to MeterReaders
    return problem

@pytest.fixture
def energyhub_df():
    df = pd.DataFrame(np.array([[200, 600],[600, 200]]),columns= ['p1','p2'])
    return df

@pytest.fixture
def energyplus_df():
    df = EPdf = pd.DataFrame(np.array([[0.5],[0.8]]),columns= ['p1'])
    return df

@pytest.fixture
def hub_problem():

    parameters = [['LINEAR_CAPITAL_COSTS','Boiler'],['LINEAR_CAPITAL_COSTS','CHP']]
    objectives = ['total_cost','total_carbon']

    problem = EHProblem(parameters, objectives)
    return problem

def test_evaluatorEP(building, problem):
    """To make sure EvaluatorEP can be initialised and works as intended"""

    evaluator = EvaluatorEP(problem, building)
    result = evaluator([0.5]) # run with thickness set to 0.5

    assert np.isclose(result[0], 1818735943.9307632) and np.isclose(result[1], 2172045529.871896), f'Unexpected result for EvaluatorEP, {result}'
    #change this to 0 to see stdout and stderr
    assert 1


def test_evaluatorEH_single(hub, hub_problem):
    """To make sure EvaluatorEH can be initialised and works as intended"""
    
    evaluator = EvaluatorEH(hub_problem,hub)
    result = evaluator([200, 600])

    assert np.isclose(result[0], 1256.12) and np.isclose(result[1], 53.2299), f'Unexpected result for EvaluatorEH, {result}'
    #change this to 0 to see stdout and stderr
    assert 1

def test_evaluatorEH_df(hub, hub_problem, energyhub_df):
    """To make sure EvaluatorEH df_apply works as intended"""
    
    evaluator = EvaluatorEH(hub_problem,hub)
    result = evaluator.df_apply(energyhub_df)

    assert np.isclose(result.iat[0,0], 1256.12) and np.isclose(result.iat[0,1], 53.2299) and np.isclose(result.iat[1,0], 1527.39) and np.isclose(result.iat[1,1], 33.7551) , f'Unexpected result for EvaluatorEH, {result}'
    #change this to 0 to see stdout and stderr
    assert 1

def test_evaluatorEH_EP(hub, hub_problem, building, problem):
    """To make sure that base EvaluatorEP output can be used in an EvaluatorEH"""
    evaluatorEP = EvaluatorEP(problem,building)
    evaluatorEH = EvaluatorEH(hub_problem,hub)
    result = evaluatorEH(evaluatorEP([0.5]))

    assert np.isclose(result[0], 2721700000.0) and np.isclose(result[1], 33.7551), f'Unexpected result for EvaluatorEH, {result}'
    #change this to 0 to see stdout and stderr
    assert 1

def test_evaluatorEH_TS(hub):
    """To make sure that EvaluatorEH can accept a time series as input"""
    timeseries = [{0: 4.0, 1: 8.0, 2: 6.0, 3: 5.0, 4: 7.0, 5: 7.0, 6: 7.0, 7: 7.0, 8: 7.0, 9: 7.0, 10: 7.0}]
    TimeSeries_parameters = [['LOADS','Elec']]
    objectives = ['total_cost','total_carbon']
    TimeSeries_problem = EHProblem(TimeSeries_parameters, objectives)
    evaluatorEH = EvaluatorEH(TimeSeries_problem, hub)
    result = evaluatorEH(timeseries)

    assert np.isclose(result[0], 1844.18) and np.isclose(result[1], 44.6054), f'Unexpected result for EvaluatorEH, {result}'
    #change this to 0 to see stdout and stderr
    assert 1

def test_evaluatorEH_TS_df(hub):
    """To make sure that EvaluatorEH's df_apply can accept a dataframe of time series as input"""
    default_timeseries = [{0: 1.0, 1: 4.0, 2: 4.0, 3: 4.0, 4: 4.0, 5: 4.0, 6: 4.0, 7: 4.0, 8: 4.0, 9: 4.0, 10: 4.0}, \
        {0: 20.0, 1: 20.0, 2: 20.0, 3: 20.0, 4: 20.0, 5: 20.0, 6: 20.0, 7: 12.0, 8: 12.0, 9: 12.0, 10: 12.0}]
    just_modified_heat = [{0: 1.0, 1: 4.0, 2: 4.0, 3: 4.0, 4: 4.0, 5: 4.0, 6: 4.0, 7: 4.0, 8: 4.0, 9: 4.0, 10: 4.0}, \
        {0: 18.0, 1: 18.0, 2: 18.0, 3: 18.0, 4: 18.0, 5: 18.0, 6: 18.0, 7: 16.0, 8: 16.0, 9: 16.0, 10: 16.0}]
    just_modified_elec = [{0: 4.0, 1: 8.0, 2: 6.0, 3: 5.0, 4: 7.0, 5: 7.0, 6: 7.0, 7: 7.0, 8: 7.0, 9: 7.0, 10: 7.0}, \
        {0: 20.0, 1: 20.0, 2: 20.0, 3: 20.0, 4: 20.0, 5: 20.0, 6: 20.0, 7: 12.0, 8: 12.0, 9: 12.0, 10: 12.0}]
    modified_timeseries = [{0: 4.0, 1: 8.0, 2: 6.0, 3: 5.0, 4: 7.0, 5: 7.0, 6: 7.0, 7: 7.0, 8: 7.0, 9: 7.0, 10: 7.0}, \
        {0: 18.0, 1: 18.0, 2: 18.0, 3: 18.0, 4: 18.0, 5: 18.0, 6: 18.0, 7: 16.0, 8: 16.0, 9: 16.0, 10: 16.0}]
    timeseries_df = pd.DataFrame(np.array([default_timeseries,just_modified_heat, just_modified_elec, modified_timeseries]),columns= ['p1','p2'])

    TSDFparameters = [['LOADS','Elec'],['LOADS','Heat']]
    TSDFobjectives = ['total_cost','total_carbon']
    TSDFproblem = EHProblem(TSDFparameters,TSDFobjectives)

    evalutorEH = EvaluatorEH(TSDFproblem, hub)
    result = evalutorEH.df_apply(timeseries_df)
    
    assert np.isclose(result.iat[0,0], 1846.19) and np.isclose(result.iat[0,1], 33.7551) \
        and np.isclose(result.iat[1,0], 1850.01) and np.isclose(result.iat[1,1], 33.7190) \
        and np.isclose(result.iat[2,0], 1844.18) and np.isclose(result.iat[2,1], 44.6054) \
        and np.isclose(result.iat[3,0], 1847.10) and np.isclose(result.iat[3,1], 44.5693), f'Unexpected result for EvaluatorEH, {result}'
    #change this to 0 to see stdout and stderr
    assert 1

def test_evaluatorEH_EP_df(hub, hub_problem, building, problem, energyplus_df):
    """To make sure that dataframe EvaluatorEP output can be used in an EvaluatorEH"""
    evaluatorEP = EvaluatorEP(problem,building)
    evaluatorEH = EvaluatorEH(hub_problem,hub)
    result = evaluatorEH.df_apply(evaluatorEP.df_apply(energyplus_df))

    assert np.isclose(result.iat[0,0], 2.721700e+09) and np.isclose(result.iat[0,1], 33.7551) and np.isclose(result.iat[1,0], 2.705480e+09) and np.isclose(result.iat[1,1], 33.7551) , f'Unexpected result for EvaluatorEH, {result}'
    #change this to 0 to see stdout and stderr
    assert 1

def test_evaluatorSR(building, problem):
    """To make sure EvaluatorSR can be initialised and works as intended"""
    def function(values):
        return ((values[0], values[0]**2), ())

    # this denotes a problem which takes 1 input, produces 2 outputs and no constraints. The placeholder parameters/objectives will be generated automatically.
    new_problem = Problem(1,2,0)

    evaluator_1 = EvaluatorSR(function, problem)
    evaluator_2 = EvaluatorSR(function, new_problem)
    result_1 = evaluator_1([4])
    result_2 = evaluator_2([4])

    assert result_1 == (4, 16), f'Unexpected result for EvaluatorSR with EPProblem, {result_1}'
    assert result_2 == (4, 16), f'Unexpected result for EvaluatorSR with custom problem, {result_2}'

    #change this to 0 to see stdout and stderr
    assert 1
